from functools import wraps
from warnings import warn


def leave_unchanged(v):
    return v


def deprecated_arugment(old, new, converter=leave_unchanged):
    """Decorator to convert an old argument to a new one.

    This decorator will convert an old argument to a new one. It will also
    issue a deprecation warning. By default, the old argument will be passed
    through unchanged. If a converter function is provided, it will be used to
    convert the old argument to the new one.

    Args:
        old: The old argument name.
        new: The new argument name.
        converter: A function to convert the old argument to the new one.
    """

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if old in kwargs:
                kwargs[new] = converter(kwargs[old])
                del kwargs[old]
                warn(
                    f"Argument {old} is deprecated. Use {new} instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )

            return f(*args, **kwargs)

        return wrapper

    return decorator



import re
from dataclasses import asdict, dataclass
from typing import Any, List, Optional, Set

from ...file_io import LoadsFromYaml, SavesToYaml
from ...subclass_registry import SubclassRegistry
from ..state import GraphObjectSchema

# TODO: Future enhancements w/ new operations or changing old ones:
#
#   1.) Dropping a part of a key to a node or relationship:
#
#       This is hard because of the case (which is fairly likely) that
#       the removal of the key takes a node or relationship from being
#       unique to no longer being unique. How do we cope with that?
#       Do we merge the nodes together? Do we make the user drop the data
#       and reingest it with these new rules? There are multiple
#       possibilities here that require both design and consideration
#       that we want to defer for now.
#
#   2.) Providing defaults dynamically:
#
#       In some ways related to (1), this is hard because we don't have
#       a formal of a data model coming back from the database.
#       This ins't an ORM and it seems  that to be able to do this in
#       both a database-agnostic and intuitive way, we'd need to almost
#       build that in such a way.
#

# We need to ignore overrides because we want to be able to
# use `dataclass` to define the operations and then register
# them with the registry. If we don't ignore overrides, then
# we'll get an error when we try to register the same class
# twice (Which happens when we use `dataclass` as a decorator)
OPERATION_SUBCLASS_REGISTRY = SubclassRegistry(ignore_overrides=True)
CAMEL_TO_SNAKE_REGEX = re.compile(r"(?<!^)(?=[A-Z])")


def camel_case_to_snake_case(camel_case: str) -> str:
    """Convert a string from camel case to snake case.

    Args:
        camel_case: The camel case string.

    Returns:
        The snake case string.
    """
    return CAMEL_TO_SNAKE_REGEX.sub("_", camel_case).lower()


@OPERATION_SUBCLASS_REGISTRY.connect_baseclass
class Operation(LoadsFromYaml, SavesToYaml):
    """Base class for all operations.

    Operations are used to migrate the database schema from one
    version to another. Each type of operation is responsible for
    migrating a specific change to the schema.
    """

    def type_as_snake_case(self) -> str:
        operation_type_name = type(self).__name__
        return camel_case_to_snake_case(operation_type_name)

    def suggest_migration_name_slug(self) -> str:
        """Suggest a migration name slug for this operation.

        Returns:
            A migration name slug.
        """
        return "_".join(self.describe().lower().split(" "))

    def describe(self) -> str:
        return self.type_as_snake_case()

    @classmethod
    def from_file_data(cls, data):
        op_type = data["operation"]
        arguments = data["arguments"]
        operation_cls = OPERATION_SUBCLASS_REGISTRY.get(op_type)
        return operation_cls(**arguments)

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Schema

        return Schema({"operation": str, "arguments": dict})

    def to_file_data(self):
        return {
            "operation": OPERATION_SUBCLASS_REGISTRY.name_for(self.__class__),
            "arguments": asdict(self),
        }

    def reduce(self, other: "Operation") -> List["Operation"]:
        """Reduce this operation with another operation.

        This method should be implemented by subclasses to reduce this
        operation with another operation. The result of this operation should
        be a list of operations that are equivalent to applying this operation
        and then the other operation. If the operations cannot be reduced, then
        this method should return both operations in a list.
        """
        return [self] if self == other else [self, other]

    @classmethod
    def optimize(cls, operations: List["Operation"]) -> List["Operation"]:
        """Optimize a list of operations.

        This method should be implemented by subclasses to optimize a list of
        operations. The result of this operation should be a list of operations
        that are equivalent to applying all of the operations in the input list.
        """
        return OperationReducer(operations).reduce()


class OperationReducer:
    """A class that can reduce a list of operations."""

    def __init__(self, operations: List[Operation]):
        self.operations = operations

    def reduce(self) -> List[Operation]:
        """Reduce the list of operations.

        Returns:
            A list of operations that are equivalent to applying all of the
            operations in the input list.
        """
        current = self.operations
        while True:
            reduced = self.reduce_once(current)
            if reduced == current:
                break
            current = reduced
        return current

    def reduce_once(self, operations: List[Operation]) -> List[Operation]:
        # We are going to go from back to front and
        # reduce exactly one operation per iteration.
        for i, operation in enumerate(reversed(operations)):
            others = operations[: len(operations) - i - 1]
            for other in reversed(others):
                reduced = operation.reduce(other)
                if len(reduced) < 2:
                    unchanged = [o for o in operations if o not in (operation, other)]
                    return unchanged + reduced

        return operations


@dataclass(frozen=True, slots=True)
class CreateNodeType(Operation):
    """Create a node of a given type.

    The underlying database connector should use the keys to
    create a uniqueness constraint or index. The keys are used
    to determine if a node already exists.

    Attributes:
        name: The name of the node type.
        keys: The keys of the node type.
        properties: The properties of the node type.
    """

    name: str
    keys: List[str]
    properties: List[str]

    @property
    def proposed_index_name(self) -> str:
        return f"{self.name}_node_key"

    def describe(self) -> str:
        return f"Create node type {self.name}"

    def as_node_type(self) -> GraphObjectSchema:
        schema = GraphObjectSchema(self.name)
        schema.add_keys(self.keys)
        schema.add_properties(self.properties)
        return schema

    def reduce(self, other: Operation) -> List[Operation]:
        if isinstance(other, DropNodeType) and other.name == self.name:
            return []
        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class CreateRelationshipType(Operation):
    """Create a relationship of a given type.

    The underlying database connector should use the keys to
    create a uniqueness constraint or index. The keys are used
    to determine if a relationship already exists (between two nodes).

    Attributes:
        name: The name of the relationship type.
        keys: The keys of the relationship type.
        properties: The properties of the relationship type.
    """

    name: str
    keys: Set[str]
    properties: Set[str]

    @property
    def proposed_index_name(self) -> str:
        return f"{self.name}_relationship_key"

    def describe(self) -> str:
        return f"Create relationship type {self.name}"

    def as_relationship_type(self) -> GraphObjectSchema:
        schema = GraphObjectSchema(self.name)
        schema.add_keys(self.keys)
        schema.add_properties(self.properties)
        return schema

    def reduce(self, other: Operation) -> List[Operation]:
        if isinstance(other, DropRelationshipType) and other.name == self.name:
            return []
        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class DropNodeType(Operation):
    """Drop a node type.

    The underlying database connector should drop the node type
    and all nodes of that type. The database connector should
    also drop any relationships that are connected to the nodes
    of that type. The database connector should also drop any
    related indexes or constraints.

    Attributes:
        name: The name of the node type.
    """

    name: str

    @property
    def proposed_index_name(self) -> str:
        return f"{self.name}_node_key"

    def describe(self) -> str:
        return f"Drop node type {self.name}"

    def reduce(self, other: Operation) -> List[Operation]:
        if isinstance(other, CreateNodeType) and other.name == self.name:
            return []

        if (
            isinstance(other, (AddNodeProperty, DropNodeProperty, RenameNodeProperty))
            and other.node_type == self.name
        ):
            return [self]

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class DropRelationshipType(Operation):
    """Drop a relationship type.

    The underlying database connector should drop the relationship
    type and all relationships of that type. The database connector
    should also drop any related indexes or constraints.

    Attributes:
        name: The name of the relationship type.
    """

    name: str

    def describe(self) -> str:
        return f"Drop relationship {self.name}"

    def reduce(self, other: Operation) -> List[Operation]:
        if isinstance(other, CreateRelationshipType) and other.name == self.name:
            return []

        if (
            isinstance(
                other,
                (
                    AddRelationshipProperty,
                    DropRelationshipProperty,
                    RenameRelationshipProperty,
                ),
            )
            and other.relationship_type == self.name
        ):
            return [self]

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class RenameNodeProperty(Operation):
    """Rename a node property.

    The underlying database connector should rename the property on all nodes
    of the given type. The database connector should also rename any related
    indexes or constraints.

    Attributes:
        node_type: The type of the node.
        old_property_name: The old name of the property that is being renamed.
        new_property_name: The new name of the property that is being renamed.
    """

    node_type: str
    old_property_name: str
    new_property_name: str

    def describe(self) -> str:
        return f"Rename node property {self.old_property_name} to {self.new_property_name} on node type {self.node_type}"

    def reduce(self, other: Operation) -> List[Operation]:
        # If we add and then rename, we can just add with the new name.
        # if we rename and then drop, we can just drop the old name.
        if (
            isinstance(other, AddNodeProperty)
            and other.node_type == self.node_type
            and other.property_name == self.old_property_name
        ):
            return [
                AddNodeProperty(self.node_type, self.new_property_name, other.default)
            ]
        if (
            isinstance(other, DropNodeProperty)
            and other.node_type == self.node_type
            and other.property_name == self.old_property_name
        ):
            return [DropNodeProperty(self.node_type, self.old_property_name)]

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class RenameRelationshipProperty(Operation):
    """Rename a relationship property.

    The underlying database connector should rename the property on all
    relationships of the given type. The database connector should also
    rename any related indexes or constraints.

    Attributes:
        relationship_type: The type of the relationship.
        old_property_name: The old name of the property that is being renamed.
        new_property_name: The new name of the property that is being renamed.
    """

    relationship_type: str
    old_property_name: str
    new_property_name: str

    def describe(self) -> str:
        return f"Rename relationship property {self.old_property_name} to {self.new_property_name} on relationship type {self.relationship_type}"

    def reduce(self, other: Operation) -> List[Operation]:
        # If we add and then rename, we can just add with the new name.
        # if we rename and then drop, we can just drop the old name.
        if (
            isinstance(other, AddRelationshipProperty)
            and other.relationship_type == self.relationship_type
            and other.property_name == self.old_property_name
        ):
            return [
                AddRelationshipProperty(
                    self.relationship_type, self.new_property_name, other.default
                )
            ]

        if (
            isinstance(other, DropRelationshipProperty)
            and other.relationship_type == self.relationship_type
            and other.property_name == self.old_property_name
        ):
            return [
                DropRelationshipProperty(self.relationship_type, self.old_property_name)
            ]

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class RenameNodeType(Operation):
    """Rename a node type.

    The underlying database connector should rename the node type and all nodes
    of that type. The database connector should also rename any related indexes
    or constraints.

    Attributes:
        old_type: The old name of the node type.
        new_type: The new name of the node type.
    """

    old_type: str
    new_type: str

    @property
    def old_proposed_index_name(self) -> str:
        return f"{self.old_type}_node_key"

    @property
    def new_proposed_index_name(self) -> str:
        return f"{self.new_type}_node_key"

    def describe(self) -> str:
        return f"Rename node type {self.old_type} to {self.new_type}"

    def reduce(self, other: Operation) -> List[Operation]:
        # If we create and then rename, we can just create with the new name.
        # If we rename and then drop, we can just drop the old name.
        if isinstance(other, CreateNodeType) and other.name == self.old_type:
            return [CreateNodeType(self.new_type, other.keys, other.properties)]

        if isinstance(other, DropNodeType) and other.name == self.new_type:
            return [DropNodeType(self.old_type)]

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class RenameRelationshipType(Operation):
    """Rename a relationship type.

    The underlying database connector should rename the relationship type and
    all relationships of that type. The database connector should also rename
    any related indexes or constraints.

    Attributes:
        old_type: The old name of the relationship type.
        new_type: The new name of the relationship type.
    """

    old_type: str
    new_type: str

    def describe(self) -> str:
        return f"Rename relationship type {self.old_type} to {self.new_type}"

    def reduce(self, other: Operation) -> List[Operation]:
        # If we create and then rename, we can just create with the new name.
        # If we rename and then drop, we can just drop the old name.
        if isinstance(other, CreateRelationshipType) and other.name == self.old_type:
            return [CreateRelationshipType(self.new_type, other.keys, other.properties)]

        if isinstance(other, DropRelationshipType) and other.name == self.new_type:
            return [DropRelationshipType(self.old_type)]

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class AddAdditionalNodePropertyIndex(Operation):
    """Add a supplemental field index.

    The underlying database connector should add an index for the given field.

    Attributes:
        node_type: The type of the node.
        field_name: The name of the field that should be indexed.
    """

    node_type: str
    field_name: str

    @property
    def proposed_index_name(self) -> str:
        return f"{self.node_type}_{self.field_name}_additional_index"

    def describe(self) -> str:
        return f"Add additional index for node type {self.node_type} on field {self.field_name}"


@dataclass(frozen=True, slots=True)
class DropAdditionalNodePropertyIndex(Operation):
    """Drop a supplemental field index.

    The underlying database connector should drop an index for the given field.

    Attributes:
        node_type: The type of the node.
        field_name: The name of the field that shouldn't be indexed.
    """

    node_type: str
    field_name: str

    @property
    def proposed_index_name(self) -> str:
        return f"{self.node_type}_{self.field_name}_additional_index"

    def describe(self) -> str:
        return f"Drop additional index for node type {self.node_type} on field {self.field_name}"


@dataclass(frozen=True, slots=True)
class AddAdditionalRelationshipPropertyIndex(Operation):
    """Add a supplemental field index.

    The underlying database connector should add an index for the given field.

    Attributes:
        relationship_type: The type of the relationship.
        field_name: The name of the field that should be indexed.
    """

    relationship_type: str
    field_name: str

    @property
    def proposed_index_name(self) -> str:
        return f"{self.relationship_type}_{self.field_name}_additional_index"

    def describe(self) -> str:
        return f"Add additional index for relationship type {self.relationship_type} on field {self.field_name}"


@dataclass(frozen=True, slots=True)
class DropAdditionalRelationshipPropertyIndex(Operation):
    """Drop a supplemental field index.

    The underlying database connector should drop an index for the given field.

    Attributes:
        relationship_type: The type of the relationship.
        field_name: The name of the field that shouldn't be indexed.
    """

    relationship_type: str
    field_name: str

    @property
    def proposed_index_name(self) -> str:
        return f"{self.relationship_type}_{self.field_name}_additional_index"

    def describe(self) -> str:
        return f"Drop additional index for relationship type {self.relationship_type} on field {self.field_name}"


@dataclass(frozen=True, slots=True)
class AddNodeProperty(Operation):
    """Add a property to a node type.

    The underlying database connector should add the property to all nodes of
    the given type. If a default is provided, the database connector should
    populate the property with the default value for all nodes of the
    given type.

    Attributes:
        node_type: The type of the node.
        property_name: The name of the property that is being added.
        default: The default value for the property.
    """

    node_type: str
    property_name: str
    default: Optional[Any] = None

    def describe(self) -> str:
        return f"Add property {self.property_name} to node type {self.node_type}"

    def reduce(self, other: Operation) -> List[Operation]:
        # If we add and then rename, we can just add with the new name.
        # If we add and then drop, we can just no-op.
        if (
            isinstance(other, RenameNodeProperty)
            and other.node_type == self.node_type
            and other.old_property_name == self.property_name
        ):
            return [
                AddNodeProperty(self.node_type, other.new_property_name, self.default)
            ]

        if (
            isinstance(other, DropNodeProperty)
            and other.node_type == self.node_type
            and other.property_name == self.property_name
        ):
            return []

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class AddRelationshipProperty(Operation):
    """Add a property to a relationship type.

    The underlying database connector should add the property to all
    relationships of the given type. If a default is provided, the database
    connector should populate the property with the default value for all
    relationships of the given type.

    Attributes:
        relationship_type: The type of the relationship.
        property_name: The name of the property that is being added.
        default: The default value for the property.
    """

    relationship_type: str
    property_name: str
    default: Optional[Any] = None

    def describe(self) -> str:
        return f"Add property {self.property_name} to relationship type {self.relationship_type}"

    def reduce(self, other: Operation) -> List[Operation]:
        # If we add and then rename, we can just add with the new name.
        # If we add and then drop, we can just no-op.
        if (
            isinstance(other, RenameRelationshipProperty)
            and other.relationship_type == self.relationship_type
            and other.old_property_name == self.property_name
        ):
            return [
                AddRelationshipProperty(
                    self.relationship_type, other.new_property_name, self.default
                )
            ]

        if (
            isinstance(other, DropRelationshipProperty)
            and other.relationship_type == self.relationship_type
            and other.property_name == self.property_name
        ):
            return []

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class DropNodeProperty(Operation):
    """Drop a property from a node type.

    The underlying database connector should drop the property from all nodes
    of the given type.

    Attributes:
        node_type: The type of the node.
        property_name: The name of the property that is being dropped.
    """

    node_type: str
    property_name: str

    def describe(self) -> str:
        return f"Drop property {self.property_name} from node type {self.node_type}"

    def reduce(self, other: Operation) -> List[Operation]:
        # If we add and then drop, we can just no-op.
        # If we rename and then drop, we can just drop the original name.
        if (
            isinstance(other, AddNodeProperty)
            and other.node_type == self.node_type
            and other.property_name == self.property_name
        ):
            return []

        if (
            isinstance(other, RenameNodeProperty)
            and other.node_type == self.node_type
            and other.old_property_name == self.property_name
        ):
            return [self]

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class DropRelationshipProperty(Operation):
    """Drop a property from a relationship type.

    The underlying database connector should drop the property from all
    relationships of the given type.

    Attributes:
        relationship_type: The type of the relationship.
        property_name: The name of the property that is being dropped.
    """

    relationship_type: str
    property_name: str

    def describe(self) -> str:
        return f"Drop property {self.property_name} from relationship type {self.relationship_type}"

    def reduce(self, other: Operation) -> List[Operation]:
        # If we add and then drop, we can just no-op.
        # If we rename and then drop, we can just drop the original name.
        if (
            isinstance(other, AddRelationshipProperty)
            and other.relationship_type == self.relationship_type
            and other.property_name == self.property_name
        ):
            return []

        if (
            isinstance(other, RenameRelationshipProperty)
            and other.relationship_type == self.relationship_type
            and other.old_property_name == self.property_name
        ):
            return [self]

        return Operation.reduce(self, other)


@dataclass(frozen=True, slots=True)
class NodeKeyExtended(Operation):
    """Add a element to a node key.

    The underlying database connector should add the key to all nodes of the
    given type. Adefault will be provided, so the database connector should
    populate the key with the default value for all nodes of the given type.

    Attributes:
        node_type: The type of the node.
        added_key_property: The name of the property that is being added.
        default: The default value for the property.
    """

    node_type: str
    added_key_property: str
    default: Any

    @property
    def proposed_index_name(self) -> str:
        return f"{self.node_type}_node_key"

    def describe(self) -> str:
        return f"Extend key {self.added_key_property} on node type {self.node_type}"


@dataclass(frozen=True, slots=True)
class RelationshipKeyExtended(Operation):
    """Add a element to a relationship key.

    The underlying database connector should add the key to all relationships
    of the given type. A default will be provided, so the database connector
    should populate the key with the default value for all relationships of
    the given type.

    Attributes:
        relationship_type: The type of the relationship.
        added_key_property: The name of the property that is being added.
        default: The default value for the property.
    """

    relationship_type: str
    added_key_property: str
    default: Any

    def describe(self) -> str:
        return f"Extend key {self.added_key_property} on relationship type {self.relationship_type}"


@dataclass(frozen=True, slots=True)
class NodeKeyPartRenamed(Operation):
    """Rename a part of a node key.

    The underlying database connector should rename the key part on all nodes
    of the given type. The database connector should also rename any related
    indexes or constraints.

    Attributes:
        node_type: The type of the node.
        old_key_part_name: The old name of the property that is being renamed.
        new_key_part_name: The new name of the property that is being renamed.
    """

    node_type: str
    old_key_part_name: str
    new_key_part_name: str

    @property
    def proposed_index_name(self) -> str:
        return f"{self.node_type}_node_key"

    def describe(self) -> str:
        return f"Rename key part {self.old_key_part_name} to {self.new_key_part_name} on node type {self.node_type}"


@dataclass(frozen=True, slots=True)
class RelationshipKeyPartRenamed(Operation):
    """Rename a part of a relationship key.

    The underlying database connector should rename the key part on all
    relationships of the given type. The database connector should also rename
    any related indexes or constraints.

    Attributes:
        relationship_type: The type of the relationship.
        old_key_part_name: The old name of the property that is being renamed.
        new_key_part_name: The new name of the property that is being renamed.
    """

    relationship_type: str
    old_key_part_name: str
    new_key_part_name: str

    def describe(self) -> str:
        return f"Rename key part {self.old_key_part_name} to {self.new_key_part_name} on relationship type {self.relationship_type}"



from datetime import datetime
from typing import List, Optional

from .auto_change_detector import AutoChangeDetector
from .migrations import Migration, MigrationGraph
from .operations import Operation


class AutoMigrationMaker:
    """A class that can make migrations automatically.

    This class is responsible for detecting changes to the schema and
    automatically generating migrations from those changes.
    """

    __slots__ = ("migration_graph", "auto_detector")

    def __init__(
        self,
        migration_graph: MigrationGraph,
        auto_detector: AutoChangeDetector,
    ) -> None:
        """Initialize the migration maker.

        Args:
            migration_graph: The migration graph to use.
            auto_detector: The auto detector of schema changes.
        """

        self.migration_graph = migration_graph
        self.auto_detector = auto_detector

    def make_migration_name(self, operations: List[Operation]) -> str:
        """Make a migration name from a list of operations.

        If the list of operations contains only one operation, then the
        migration name will be the timestamp followed by the slug of the
        operation. Otherwise, the migration name will be the timestamp.

        Args:
            operations: A list of operations.

        Returns:
            A migration name.
        """
        base_name = datetime.now().strftime("%Y%m%d%H%M%S")
        if len(operations) == 1:
            return f"{base_name}_{operations[0].suggest_migration_name_slug()}"

        return base_name

    def get_migration_dependency_names(self) -> List[str]:
        """Get the names of the migrations that this migration depends on.

        Logically, we are building a new migration that depends on the
        current state of the database (all migrations thus far) as well as
        the current state of the schema (from all pipelines).

        Therefore, we need this migration to depend on all leaf nodes in the
        migration graph. Which is really all that this method does.

        Returns:
            A list of migration names.
        """
        return [m.name for m in self.migration_graph.get_leaf_migrations()]

    async def make_migration(self) -> Optional[Migration]:
        """Make a migration from the changes that were detected.

        Returns:
            A migration describing the changes that were detected. If no
            changes were detected, then this method returns None.
        """
        operations = await self.auto_detector.detect_changes()
        if len(operations) == 0:
            return None

        return Migration(
            name=self.make_migration_name(operations),
            operations=operations,
            dependencies=self.get_migration_dependency_names(),
        )



import re
from abc import abstractmethod
from logging import getLogger
from typing import Any, AsyncGenerator, Dict, Iterable, List, Optional

from .flush import Flush
from .object_storage import ObjectStore
from .step import PassStep, Step, StepContext
from .value_providers import ProviderContext, StaticValueOrValueProvider, ValueProvider


class Filter(Step):
    """A `Filter` takes a given record and evaluates whether or not it should continue downstream.

    `Filter` steps generally make up the middle of an ETL pipeline and are
    responsible for ensuring only relevant records make it through.
    """

    async def process_record(self, record, _) -> AsyncGenerator[object, None]:
        if record is Flush or not await self.filter_record(record):
            yield record

    @abstractmethod
    async def filter_record(self, record: Any) -> bool:
        raise NotImplementedError


class ValueMatcher:
    @classmethod
    def from_file_data(
        cls,
        value: StaticValueOrValueProvider,
        possibilities: Iterable[StaticValueOrValueProvider],
        normalization: Optional[Dict[str, Any]] = None,
    ):
        return cls(
            value_provider=ValueProvider.guarantee_value_provider(value),
            possibilities=ValueProvider.guarantee_provider_list(possibilities),
            normalization=(normalization or {}),
        )

    def __init__(
        self,
        value_provider: ValueProvider,
        possibilities: Iterable[ValueProvider],
        normalization: Dict[str, Any],
    ) -> None:
        self.value_provider = value_provider
        self.possibilities = possibilities
        self.normalization = normalization

    def does_match(self, context: ProviderContext):
        actual_value = self.value_provider.normalize_single_value(
            context, self.normalization
        )

        return any(
            possibility_provider.normalize_single_value(context, self.normalization)
            == actual_value
            for possibility_provider in self.possibilities
        )


class ValuesMatchPossibilitiesFilter(Filter):
    """A filter that checks if a given value matches any of a set of possibilities."""

    @classmethod
    def from_file_data(cls, *, fields: Iterable[Dict[str, Any]]):
        value_matchers = [ValueMatcher.from_file_data(**field) for field in fields]
        return cls(value_matchers=value_matchers)

    def __init__(self, value_matchers: Iterable[ValueMatcher]):
        self.value_matchers = value_matchers

    async def filter_record(self, item):
        context_from_record = ProviderContext(item, None)
        return not all(
            matcher.does_match(context_from_record) for matcher in self.value_matchers
        )


class ExcludeWhenValuesMatchPossibilities(Filter):
    @classmethod
    def from_file_data(cls, **kwargs):
        inner = ValuesMatchPossibilitiesFilter.from_file_data(**kwargs)
        return cls(inner)

    def __init__(self, inner: ValuesMatchPossibilitiesFilter) -> None:
        self.inner = inner

    async def filter_record(self, record: Any) -> bool:
        return not await self.inner.filter_record(record)


class RegexMatcher:
    @classmethod
    def from_file_data(
        cls,
        value: StaticValueOrValueProvider,
        regex: str,
        include: bool = True,
        normalization: Optional[Dict[str, Any]] = None,
    ):
        return cls(
            value_provider=ValueProvider.guarantee_value_provider(value),
            regex=regex,
            include=include,
            normalization=(normalization or {}),
        )

    def __init__(
        self,
        value_provider: StaticValueOrValueProvider,
        regex: StaticValueOrValueProvider,
        include: bool,
        normalization: Dict[str, Any],
    ) -> None:
        self.value_provider = value_provider
        self.regex = re.compile(regex)
        self.include = include
        self.normalization = normalization

    def should_include(self, context: ProviderContext) -> bool:
        actual_value = self.value_provider.normalize_single_value(
            context, self.normalization
        )
        match = self.regex.match(actual_value) is not None
        return not match if self.include else match


class ValueMatchesRegexFilter(Filter):
    """A filter that includes/excludes a given value based on a matched regex."""

    @classmethod
    def from_file_data(
        cls,
        value: StaticValueOrValueProvider,
        regex: StaticValueOrValueProvider,
        include: bool,
        normalize: Optional[Dict[str, Any]] = None,
    ):
        regex_matcher = RegexMatcher.from_file_data(value, regex, include, normalize)
        return cls(regex_matcher=regex_matcher)

    def __init__(self, regex_matcher: RegexMatcher):
        self.regex_matcher = regex_matcher

    async def filter_record(self, item: Any):
        context_from_record = ProviderContext(item, None)
        return self.regex_matcher.should_include(context_from_record)


try:
    import genson
    import jsonschema

    INFERRED_OBJECT_SCHEMA_KEY = "inferred_object_schema"

    class SchemaEnforcementMode:
        """A mode of schema enforcement that can transition between different states."""

        def should_filter(self, record) -> bool:
            return True  # pragma: no cover

        def start(self, object_store: ObjectStore) -> "SchemaEnforcementMode":
            return self  # pragma: no cover

        def inform_mode_change(self) -> "SchemaEnforcementMode":
            return self  # pragma: no cover

        @staticmethod
        def from_enforcement_policy(
            enforcement_policy: str, schema: "Schema"
        ) -> "SchemaEnforcementMode":
            if enforcement_policy == "enforce":
                return EnforceSchema(schema)
            elif enforcement_policy == "warn":
                return WarnSchema(schema)
            raise ValueError(f"Invalid enforcement policy: {enforcement_policy}")

    class Schema:
        """A schema that can be enforced on records."""

        def __init__(self, schema_dict):
            self.schema_dict = schema_dict
            self.validator = jsonschema.Draft7Validator(schema_dict)

        def validate(self, record) -> Optional[List[str]]:
            errors = sorted(self.validator.iter_errors(record), key=lambda e: e.path)
            return list(map(lambda e: e.message, errors))

        def persist(self, object_store: ObjectStore, key: str):
            object_store.put_picklable(key, self.schema_dict)

        @staticmethod
        def get_from_object_store(
            object_store: ObjectStore, key: str
        ) -> Optional["Schema"]:
            if schema_dict := object_store.get_pickled(key):
                return Schema(schema_dict)
            return None

    class SchemaBuilder:
        """A helper class to build a schema from records."""

        def __init__(self):
            self._inner = genson.SchemaBuilder()
            self.collected_samples = 0

        def add_record(self, record):
            self._inner.add_object(record)
            self.collected_samples += 1

        def build_schema(self) -> Schema:
            return Schema(self._inner.to_schema())

    class FetchSchema(SchemaEnforcementMode):
        """Fetches a schema from the object store and enforces it."""

        def __init__(self, key: str, enforcement_policy: str):
            self.key = key
            self.enforcement_policy = enforcement_policy

        def start(self, object_store: ObjectStore):
            if schema := Schema.get_from_object_store(object_store, self.key):
                return SchemaEnforcementMode.from_enforcement_policy(
                    self.enforcement_policy, schema
                )

            raise ValueError(f"Schema with key {self.key} not found in object store.")

    class InferSchema(SchemaEnforcementMode):
        """Infers a schema from the first `sample_size` records.

        If the schema is already present in the object store, it will switch to
        out of inferring immediately. Otherwise, it will infer the schema from the
        first `sample_size` records and then switch.

        Whether or not the switch is to ENFORCE or WARN mode is determined by
        the `warn` parameter.
        """

        def __init__(self, sample_size: int, enforcement_policy: str):
            self.sample_size = sample_size
            self.builder = SchemaBuilder()
            self.enforcement_policy = enforcement_policy
            self.logger = getLogger(self.__class__.__name__)
            self.object_store = None

        def start(self, object_store):
            self.object_store = object_store
            if schema := Schema.get_from_object_store(
                object_store, INFERRED_OBJECT_SCHEMA_KEY
            ):
                return self.switch_out(schema)
            return self

        def should_filter(self, record):
            self.builder.add_record(record)
            self.logger.debug(
                "In infer mode. Not filtering.",
                extra={"samples": self.builder.collected_samples},
            )
            return False

        def switch_out(self, schema: Schema) -> SchemaEnforcementMode:
            self.logger.info("Switching out of infer mode.")
            schema.persist(self.object_store, INFERRED_OBJECT_SCHEMA_KEY)
            return SchemaEnforcementMode.from_enforcement_policy(
                self.enforcement_policy, schema
            )

        def inform_mode_change(self):
            if self.builder.collected_samples >= self.sample_size:
                return self.switch_out(self.builder.build_schema())

            return self

    class EnforceSchema(SchemaEnforcementMode):
        """Filters records that do not match the schema.

        Once in this mode, no transition is possible. The schema is enforced
        until the end of the pipeline.
        """

        def __init__(self, schema: Schema):
            self.schema = schema
            self.logger = getLogger(self.__class__.__name__)

        def should_filter(self, record):
            if errors := self.schema.validate(record):
                self.logger.error(
                    "Schema validation failed. Filtering due to ENFORCE mode.",
                    extra={"errors": errors},
                )
            return bool(errors)

    class WarnSchema(SchemaEnforcementMode):
        """Permits (while warning about) records that do not match the schema.

        Once in this mode, no transition is possible. The schema is warned about
        until the end of the pipeline.
        """

        def __init__(self, schema: Schema):
            self.schema = schema
            self.logger = getLogger(self.__class__.__name__)

        def should_filter(self, record):
            if errors := self.schema.validate(record):
                self.logger.warning(
                    "Schema validation failed. NOT filtering due to WARN mode.",
                    extra={"errors": errors},
                )
            return False

    class SchemaEnforcer(Filter):
        """A filter that enforces a schema on records.

        This effectively works as a state machine that transitions between
        different modes of schema enforcement.
        """

        @classmethod
        def from_file_data(
            cls,
            enforcement_policy: str = "enforce",
            key: str = None,
            inference_sample_size: int = 1000,
        ):
            if key:
                return cls(FetchSchema(key, enforcement_policy))
            return cls(InferSchema(inference_sample_size, enforcement_policy))

        def __init__(self, mode: SchemaEnforcementMode):
            self.mode = mode

        async def start(self, context: StepContext):
            self.mode = self.mode.start(context.object_store)

        async def filter_record(self, record):
            self.mode = self.mode.inform_mode_change()
            return self.mode.should_filter(record)

except ImportError:

    class SchemaEnforcer(PassStep):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            raise ImportError(
                "SchemaEnforcer requires genson and jsonschema to be installed. Install the `validation` extra."
            )



import hashlib
from dataclasses import dataclass, field
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set

from ..file_io import (
    LazyLoadedArgument,
    LazyLoadedTagSafeLoader,
    LoadsFromYaml,
    LoadsFromYamlFile,
)
from .argument_resolvers import set_config
from .class_loader import ClassLoader
from .normalizers import Normalizer
from .object_storage import ObjectStore
from .pipeline import Pipeline
from .scope_config import ScopeConfig
from .step import Step
from .value_providers import ValueProvider


class InvalidPipelineDefinitionError(ValueError):
    """Raised when a pipeline definition is invalid."""

    pass


class PipelineFileSafeLoader(LazyLoadedTagSafeLoader):
    """A YAML loader that can load pipeline files.""" ""

    was_configured = False

    @classmethod
    def configure(cls):
        if cls.was_configured:
            return

        for normalizer in Normalizer.all():
            normalizer.setup()
        for value_provider in ValueProvider.all():
            value_provider.install_yaml_tag(cls)

        cls.was_configured = True


@dataclass(slots=True)
class PipelineInitializationArguments:
    """Arguments used to initialize a pipeline from a file."""

    step_outbox_size: int = 1000
    annotations: Optional[List[str]] = None
    on_effective_configuration_resolved: Optional[Callable[[List[Dict]], None]] = None
    extra_steps: Optional[List[Step]] = None
    effecitve_config_values: Optional[ScopeConfig] = None
    object_store: ObjectStore = field(default_factory=ObjectStore.null)

    @classmethod
    def for_introspection(cls):
        return cls(annotations=["introspection"])

    @classmethod
    def for_testing(cls):
        return cls(annotations=["test"])


@dataclass(frozen=True, slots=True)
class StepDefinition(LoadsFromYaml):
    implementation_path: str
    factory_method_name: Optional[str] = None
    annotations: Optional[Set[str]] = None
    arguments: Optional[Dict[str, Any]] = None

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            {
                "implementation": str,
                Optional("factory"): str,
                Optional("annotations"): [str],
                Optional("arguments"): {str: object},
            }
        )

    @classmethod
    def from_file_data(cls, data):
        return cls(
            implementation_path=data["implementation"],
            factory_method_name=data.get("factory"),
            annotations=set(data.get("annotations", [])),
            arguments=data.get("arguments"),
        )

    def should_be_loaded(self, user_annotations: Optional[Set[str]]) -> bool:
        """Determine if this step should be loaded.

        Here are the rules:
            - If the step has no annotations, it will always be loaded.
            - If the user supplies no annotations, it will always be loaded.
            - If the step has annotations, and the user supplies annotations, it will only be loaded if the step's
                annotations are a subset of the user's annotations.

        Args:
            user_annotations: The annotations supplied by the user.

        Returns:
            True if the step should be loaded, False otherwise.
        """
        if not bool(self.annotations) or not bool(user_annotations):
            return True

        return bool(self.annotations.intersection(user_annotations))

    def load_step(self) -> Step:
        arguments = LazyLoadedArgument.resolve_if_needed(self.arguments)
        return ClassLoader.instance(Step).load_class(
            implementation=self.implementation_path,
            arguments=arguments,
            factory=self.factory_method_name,
        )


class PipelineFileContents(LoadsFromYamlFile):
    @classmethod
    def describe_yaml_schema(cls):
        from schema import Schema

        return Schema([StepDefinition.describe_yaml_schema()])

    @classmethod
    def get_loader(cls):
        PipelineFileSafeLoader.configure()
        return PipelineFileSafeLoader

    @classmethod
    def from_file_data(cls, data: List[Dict]):
        step_definitions = [
            StepDefinition.from_file_data(definition_data) for definition_data in data
        ]
        return cls(step_definitions)

    def __init__(self, step_definitions: List[StepDefinition]) -> None:
        self.step_definitions = step_definitions

    def initialize_with_arguments(self, init_args: PipelineInitializationArguments):
        with set_config(init_args.effecitve_config_values):
            steps_defined_in_file = [
                step_definition.load_step()
                for step_definition in self.step_definitions
                if step_definition.should_be_loaded(init_args.annotations)
            ]
            steps = steps_defined_in_file + (init_args.extra_steps or [])
        return Pipeline(
            steps,
            step_outbox_size=init_args.step_outbox_size,
            object_store=init_args.object_store,
        )


class PipelineFile:
    def __init__(self, file_path: Path):
        self.file_path = file_path
        self.logger = getLogger(self.__class__.__name__)

    def file_sha_256(self) -> str:
        sha = hashlib.sha256()
        b = bytearray(128 * 1024)
        mv = memoryview(b)
        with self.file_path.open("rb", buffering=0) as file:
            while n := file.readinto(mv):
                sha.update(mv[:n])
        return sha.hexdigest()

    def load_pipeline(
        self, init_args: Optional[PipelineInitializationArguments] = None
    ) -> Pipeline:
        self.logger.info("Loading Pipeline")
        init_args = init_args or PipelineInitializationArguments()
        init_args.object_store = init_args.object_store.namespaced(self.file_sha_256())
        contents = self.get_contents()
        return contents.initialize_with_arguments(init_args)

    def get_contents(self) -> PipelineFileContents:
        return PipelineFileContents.read_from_file(self.file_path)



from abc import ABC, abstractmethod

from ...pipeline.value_providers import (
    ProviderContext,
    StaticValueOrValueProvider,
    ValueProvider,
)
from ...subclass_registry import SubclassRegistry

CONDITION_SUBCLASS_REGISTRY = SubclassRegistry()
COMPARISON_OPERATOR_SUBCLASS_REGISTRY = SubclassRegistry()


@CONDITION_SUBCLASS_REGISTRY.connect_baseclass
class Condition(ABC):
    @abstractmethod
    def evaluate(self, context: ProviderContext) -> bool:
        pass

    @classmethod
    @abstractmethod
    def from_file_data(cls, **arguments) -> "Condition":
        pass

    @staticmethod
    def from_tagged_file_data(**data) -> "Condition":
        name = data.pop("type", CONDITION_SUBCLASS_REGISTRY.name_for(AlwaysTrue))
        class_to_load = CONDITION_SUBCLASS_REGISTRY.get(name)
        return class_to_load.from_file_data(**data)


class AlwaysTrue(Condition, alias="true"):
    def evaluate(self, context: ProviderContext) -> bool:
        return True

    @classmethod
    def from_file_data(cls, **arguments) -> "Condition":
        return cls()


class Or(Condition, alias="or"):
    def __init__(self, *conditions: Condition):
        self.conditions = conditions

    def evaluate(self, context: ProviderContext) -> bool:
        return any(condition.evaluate(context) for condition in self.conditions)

    @classmethod
    def from_file_data(cls, **arguments) -> "Condition":
        conditions = [
            Condition.from_tagged_file_data(**condition)
            for condition in arguments["conditions"]
        ]
        return cls(*conditions)


class And(Condition, alias="and"):
    def __init__(self, *conditions: Condition):
        self.conditions = conditions

    def evaluate(self, context: ProviderContext) -> bool:
        return all(condition.evaluate(context) for condition in self.conditions)

    @classmethod
    def from_file_data(cls, **arguments) -> "Condition":
        conditions = [
            Condition.from_tagged_file_data(**condition)
            for condition in arguments["conditions"]
        ]
        return cls(*conditions)


class Not(Condition, alias="not"):
    def __init__(self, condition: Condition):
        self.condition = condition

    def evaluate(self, context: ProviderContext) -> bool:
        return not self.condition.evaluate(context)

    @classmethod
    def from_file_data(cls, **arguments) -> "Condition":
        condition = Condition.from_tagged_file_data(**arguments["condition"])
        return cls(condition)


@COMPARISON_OPERATOR_SUBCLASS_REGISTRY.connect_baseclass
class ComparisonOperator(ABC):
    @abstractmethod
    def operate(self, left, right) -> bool:
        pass


class EqualsOperator(ComparisonOperator, alias="equals"):
    def operate(self, left, right) -> bool:
        return left == right


class GreaterThanOperator(ComparisonOperator, alias="greater_than"):
    def operate(self, left, right) -> bool:
        return left > right


class LessThanOperator(ComparisonOperator, alias="less_than"):
    def operate(self, left, right) -> bool:
        return left < right


class ContainsOperator(ComparisonOperator, alias="contains"):
    def operate(self, left, right) -> bool:
        return right in left


class Comparator(Condition, alias="compare"):
    def __init__(
        self,
        left: StaticValueOrValueProvider,
        right: StaticValueOrValueProvider,
        operator: ComparisonOperator,
    ):
        self.left = ValueProvider.guarantee_value_provider(left)
        self.right = ValueProvider.guarantee_value_provider(right)
        self.operator = operator

    def evaluate(self, context: ProviderContext) -> bool:
        left = self.left.single_value(context)
        right = self.right.single_value(context)
        return self.operator.operate(left, right)

    @classmethod
    def from_file_data(cls, **arguments) -> "Condition":
        left = arguments["left"]
        right = arguments["right"]
        operator = COMPARISON_OPERATOR_SUBCLASS_REGISTRY.get(arguments["operator"])
        return cls(left=left, right=right, operator=operator())



from abc import abstractmethod
from typing import Any, AsyncGenerator, Generic, TypeVar

from ..step import Step, StepContext

R = TypeVar("R")
T = TypeVar("T")
CHECKPOINT_OBJECT_KEY = "extractor_progress_checkpoint"


class Extractor(Step, Generic[R, T]):
    """Extractors represent the source of a set of records.

    They are like any other step. However, they ignore the incoming record '
    stream and instead produce their own stream of records. For this reason
    they generally should only be set at the beginning of a pipeline.
    """

    CHECKPOINT_INTERVAL = 1000

    async def start(self, context: StepContext):
        if checkpoint := context.object_store.get_pickled(CHECKPOINT_OBJECT_KEY):
            context.info("Found Checkpoint For Extractor. Signaling to resume from it.")
            await self.resume_from_checkpoint(checkpoint)

    async def finish(self, context: StepContext):
        if not context.pipeline_encountered_fatal_error:
            context.debug(
                "Clearing checkpoint for extractor since extractor is finished."
            )
            context.object_store.delete(CHECKPOINT_OBJECT_KEY)

    async def make_checkpoint(self) -> T:
        return None

    async def resume_from_checkpoint(self, checkpoint_object: T):
        pass

    async def commit_checkpoint(self, context: StepContext) -> None:
        if checkpoint := await self.make_checkpoint():
            context.object_store.put_picklable(CHECKPOINT_OBJECT_KEY, checkpoint)

    async def emit_outstanding_records(
        self, context: StepContext
    ) -> AsyncGenerator[R, None]:
        items_generated = 0
        async for record in self.extract_records():
            yield record
            items_generated += 1
            if items_generated % self.CHECKPOINT_INTERVAL == 0:
                await self.commit_checkpoint(context)

    @abstractmethod
    def extract_records(self) -> AsyncGenerator[R, Any]:
        raise NotImplementedError



from ...schema.migrations import ProjectMigrations
from .operation import NodestreamCommand, Operation


class GenerateSquashedMigration(Operation):
    def __init__(
        self,
        migrations: ProjectMigrations,
        from_migration_name: str,
        to_migration_name: str,
    ) -> None:
        self.migrations = migrations
        self.from_migration_name = from_migration_name
        self.to_migration_name = to_migration_name

    async def perform(self, command: NodestreamCommand):
        from_migration = self.migrations.graph.get_migration(self.from_migration_name)
        to_migration = (
            self.migrations.graph.get_migration(self.to_migration_name)
            if self.to_migration_name
            else None
        )
        migration, path = self.migrations.create_squash_between(
            from_migration, to_migration
        )
        command.line(f"Generated squashed migration {migration.name}.")
        command.line(
            f"The migration contains {len(migration.operations)} schema changes."
        )
        for operation in migration.operations:
            command.line(f"  - {operation.describe()}")
        command.line(f"Migration written to {path}")
        command.line("Run `nodestream migrations run` to apply the migration.")



import asyncio
from logging import getLogger
from typing import Any, AsyncGenerator, Iterable, Optional

from nodestream.pipeline.extractors.queues.extractor import QueueConnector

from ..credential_utils import AwsClientFactory


class SQSQueueConnector(QueueConnector, alias="sqs"):
    @classmethod
    def from_file_data(
        cls,
        queue_url: str,
        message_system_attribute_names: Optional[list[str]] = ["All"],
        message_attribute_names: Optional[list[str]] = ["All"],
        max_batch_size: int = 10,
        delete_after_read: bool = True,
        max_batches: int = 10,
        **aws_client_args,
    ):
        return cls(
            queue_url=queue_url,
            message_system_attribute_names=message_system_attribute_names,
            message_attribute_names=message_attribute_names,
            max_batch_size=max_batch_size,
            delete_after_read=delete_after_read,
            max_batches=max_batches,
            sqs_client=AwsClientFactory(**aws_client_args).make_client("sqs"),
        )

    def __init__(
        self,
        sqs_client,
        queue_url: str,
        message_system_attribute_names: Optional[list[str]] = ["All"],
        message_attribute_names: Optional[list[str]] = ["All"],
        max_batch_size: int = 10,
        delete_after_read: bool = True,
        max_batches: int = 10,
    ) -> None:
        self.queue_url = queue_url
        self.message_system_attribute_names = message_system_attribute_names
        self.message_attribute_names = message_attribute_names
        self.max_batch_size = max_batch_size
        self.delete_after_read = delete_after_read
        self.max_batches = max_batches
        self.sqs_client = sqs_client
        self.logger = getLogger(__name__)

    async def poll(self) -> Iterable[Any]:
        results = []
        for _ in range(self.max_batches):
            try:
                messages = await self.get_next_messsage_batch()
                if messages is None:
                    self.logger.debug("Polling returned no messages")
                    continue
                for result in messages:
                    results.append(result)
            except Exception as e:
                self.logger.exception(f"error while polling SQS messages: {e}")
                break
        return results

    async def get_next_messsage_batch(self):
        loop = asyncio.get_running_loop()
        msgs = await loop.run_in_executor(None, self.get_message_batch)
        messages = msgs.get("Messages", [])
        if len(messages) == 0:
            return None
        return self.process_messages(messages)

    def process_messages(self, messages):
        for message in messages:
            yield message["Body"]
            if self.delete_after_read:
                self.delete_message(message)

    def get_message_batch(self) -> AsyncGenerator[Any, Any]:
        return self.sqs_client.receive_message(
            QueueUrl=self.queue_url,
            MessageSystemAttributeNames=self.message_system_attribute_names,
            MessageAttributeNames=self.message_attribute_names,
            MaxNumberOfMessages=self.max_batch_size,
        )

    def delete_message(self, msg):
        receipt_handle = msg.get("ReceiptHandle")
        return self.sqs_client.delete_message(
            QueueUrl=self.queue_url, ReceiptHandle=receipt_handle
        )



from abc import ABC, abstractmethod
from copy import deepcopy
from os import environ
from typing import Iterable

from nodestream.schema.state import SchemaExpansionCoordinator

from ..pipeline.value_providers import ProviderContext
from ..schema import ExpandsSchema, ExpandsSchemaFromChildren
from .interpretations import Interpretation

COMPLETENESS_ERROR_MESSAGE = "Each Interpreter Pass in the Multi Sequence interpreter pass must define a source node."
UNIQUENESS_ERROR_MESSAGE = (
    "Only one interpretation can generate a source node within an interpretation pass."
)
VALIDATION_FLAG = "VALIDATE_PIPELINE_SCHEMA"


class InterpretationPassError(Exception):
    pass


class InterpretationPass(ExpandsSchema, ABC):
    @classmethod
    def from_file_data(self, args):
        if args is None:
            return NullInterpretationPass()

        if len(args) > 0 and isinstance(args[0], list):
            return MultiSequenceInterpretationPass.from_file_data(args)
        return SingleSequenceInterpretationPass.from_file_data(args)

    @abstractmethod
    def apply_interpretations(self, context: ProviderContext):
        pass

    @property
    def assigns_source_nodes(self) -> bool:
        pass


class NullInterpretationPass(InterpretationPass):
    def apply_interpretations(self, context: ProviderContext):
        yield context

    @property
    def assigns_source_nodes(self) -> bool:
        return False


class SingleSequenceInterpretationPass(InterpretationPass):
    __slots__ = ("interpretations",)

    @classmethod
    def from_file_data(cls, interpretation_arg_list):
        interpretations = (
            Interpretation.from_file_data(**args) for args in interpretation_arg_list
        )
        return_class = cls(*interpretations)
        return_class.verify_uniqueness()
        return return_class

    def __init__(self, *interpretations: Interpretation):
        self.interpretations = interpretations

    # Verifies that there is only one source node creator in all the interpretations
    def verify_uniqueness(self):
        source_node_generator_count = 0
        for interpretation in self.interpretations:
            if interpretation.assigns_source_nodes:
                source_node_generator_count += 1
        if source_node_generator_count > 1 and environ.get(VALIDATION_FLAG, False):
            raise InterpretationPassError(COMPLETENESS_ERROR_MESSAGE)

    @property
    def assigns_source_nodes(self) -> bool:
        return any(
            interpretation.assigns_source_nodes
            for interpretation in self.interpretations
        )

    # If we assign a source node at this level, add to the schema at this level.
    @property
    def exclusively_assigns_source_nodes(self) -> bool:
        return any(
            (
                interpretation.assigns_source_nodes
                and not isinstance(interpretation, ExpandsSchemaFromChildren)
            )
            for interpretation in self.interpretations
        )

    def apply_interpretations(self, context: ProviderContext):
        for interpretation in self.interpretations:
            interpretation.interpret(context)
        yield context

    # Ordering the interpretations to leave the source node creator last.
    @property
    def schema_ordered_interpretations(self):
        source_node_interpretation = None
        interpretations_copy = [
            interpretation for interpretation in self.interpretations
        ]

        for index, interpretation in enumerate(interpretations_copy):
            if interpretation.assigns_source_nodes:
                source_node_interpretation = interpretations_copy.pop(index)

        return interpretations_copy + (
            [source_node_interpretation]
            if source_node_interpretation is not None
            else []
        )

    def expand_schema(self, coordinator: SchemaExpansionCoordinator):
        for interpretation in self.schema_ordered_interpretations:
            interpretation.expand_schema(coordinator=coordinator)

        if self.exclusively_assigns_source_nodes:
            coordinator.clear_aliases()


# For each interpretationpass they maintain their own context hence, they expand schema from children.
class MultiSequenceInterpretationPass(ExpandsSchemaFromChildren, InterpretationPass):
    __slots__ = ("passes",)

    @classmethod
    def from_file_data(cls, args):
        interpretation_passes = (InterpretationPass.from_file_data(arg) for arg in args)
        return_class = cls(*interpretation_passes)
        return_class.verify_completeness()
        return return_class

    def __init__(self, *passes: InterpretationPass) -> None:
        self.passes: list[InterpretationPass] = passes

    # If any sequence assigns source nodes, all of the passes must assign source nodes.
    def verify_completeness(self):
        if (
            self.assigns_source_nodes
            and not all(
                interpretation_pass.assigns_source_nodes
                for interpretation_pass in self.passes
            )
            and environ.get(VALIDATION_FLAG, False)
        ):
            raise InterpretationPassError(COMPLETENESS_ERROR_MESSAGE)

    @property
    def assigns_source_nodes(self) -> bool:
        return any(
            interpretation_pass.assigns_source_nodes
            for interpretation_pass in self.passes
        )

    @property
    def should_be_distinct(self) -> bool:
        return self.assigns_source_nodes

    def apply_interpretations(self, context: ProviderContext):
        for interpretation_pass in self.passes:
            provided_subcontext = deepcopy(context)
            for res in interpretation_pass.apply_interpretations(provided_subcontext):
                yield res

    def get_child_expanders(self) -> Iterable[ExpandsSchema]:
        yield from self.passes



from .audit_command import AuditCommand
from .copy import Copy
from .make_migrations import MakeMigration
from .new import New
from .nodestream_command import NodestreamCommand
from .print_schema import PrintSchema
from .remove import Remove
from .run import Run
from .run_migrations import RunMigrations
from .scaffold import Scaffold
from .show import Show
from .show_migrations import ShowMigrations
from .squash_migrations import SquashMigration

__all__ = (
    "AuditCommand",
    "Copy",
    "MakeMigration",
    "New",
    "NodestreamCommand",
    "PrintSchema",
    "Remove",
    "RunMigrations",
    "Run",
    "Scaffold",
    "ShowMigrations",
    "Show",
    "SquashMigration",
)



from ...project import Target
from ...schema.migrations import ProjectMigrations
from .operation import NodestreamCommand, Operation


class ExecuteMigrations(Operation):
    def __init__(self, migrations: ProjectMigrations, target: Target):
        self.migrations = migrations
        self.target = target

    async def perform(self, command: NodestreamCommand):
        migrator = self.target.make_migrator()
        with command.spin(
            f"Executing migrations on target {self.target.name}...",
            f"Migrations executed on target {self.target.name}.",
        ) as indicator:
            async for migration in self.migrations.execute_pending(migrator):
                indicator.set_message(
                    f"Migration {migration.name} executed successfully."
                )



from os import environ
from typing import Any, Dict, Iterable, List

from ...pipeline.value_providers import (
    ProviderContext,
    StaticValueOrValueProvider,
    ValueProvider,
)
from ...schema import ExpandsSchema, ExpandsSchemaFromChildren
from ..interpretation_passes import InterpretationPass, MultiSequenceInterpretationPass
from .interpretation import Interpretation

SWITCH_COMPLETENESS_ERROR_MESSAGE = (
    "Every branch within the switch interpreter must contain a source node generator."
)
INVALID_SWITCH_ERROR_MESSAGE = (
    "Switch interpretations cannot handle multiple interpretation passes."
)
VALIDATION_FLAG = "VALIDATE_PIPELINE_SCHEMA"


class UnhandledBranchError(ValueError):
    """Raised when a branch is not handled in a switch case."""

    def __init__(self, missing_branch_value, *args: object) -> None:
        super().__init__(
            f"'{missing_branch_value}' was not matched in switch case", *args
        )


class SwitchError(Exception):
    pass


class SwitchInterpretation(Interpretation, ExpandsSchemaFromChildren, alias="switch"):
    __slots__ = (
        "switch_on",
        "interpretations",
        "default",
        "normalization",
        "fail_on_unhandled",
    )

    @staticmethod
    def guarantee_single_interpretation_pass_from_file_data(
        file_data,
    ) -> InterpretationPass:
        # Check for multiinterpretationpass.
        if isinstance(file_data, list):
            interpretation_pass = InterpretationPass.from_file_data(file_data)
        else:
            interpretation_pass = InterpretationPass.from_file_data([file_data])

        # We do not support multiple interpretation passes within the switch interpretation.
        if isinstance(interpretation_pass, MultiSequenceInterpretationPass):
            raise SwitchError(INVALID_SWITCH_ERROR_MESSAGE)

        return interpretation_pass

    def __init__(
        self,
        switch_on: StaticValueOrValueProvider,
        cases: Dict[str, Interpretation | List[Interpretation]],
        default: Interpretation | List[Interpretation] = None,
        normalization: Dict[str, Any] = None,
        fail_on_unhandled: bool = True,
    ):
        self.switch_on = ValueProvider.guarantee_value_provider(switch_on)
        self.branches = {
            field_value: self.guarantee_single_interpretation_pass_from_file_data(
                interpretation
            )
            for field_value, interpretation in cases.items()
        }
        self.default = (
            self.guarantee_single_interpretation_pass_from_file_data(default)
            if default
            else None
        )
        self.normalization = normalization or {}
        self.fail_on_unhandled = fail_on_unhandled
        self.verify_completeness()

    @property
    def all_interpretations(self) -> Iterable[Interpretation]:
        yield from self.branches.values()
        if self.default:
            yield self.default

    @property
    def assigns_source_nodes(self) -> bool:
        # If all branches have at least one interpretation that assigns source nodes,
        # then the branches are distinct and we should not merge their schemas.
        return any(branch.assigns_source_nodes for branch in self.all_interpretations)

    # There is only a distinct context between children if this interpretation assign source nodes.
    @property
    def should_be_distinct(self) -> bool:
        return self.assigns_source_nodes

    # If this interpretation assigns source nodes, and not all of the branches including the default branch assign source nodes it is incomplete.
    def verify_completeness(self):
        if (
            self.assigns_source_nodes
            and not all(
                interpretation_pass.assigns_source_nodes
                for interpretation_pass in self.all_interpretations
            )
            and environ.get(VALIDATION_FLAG, False)
        ):
            raise SwitchError(SWITCH_COMPLETENESS_ERROR_MESSAGE)

    def get_child_expanders(self) -> Iterable[ExpandsSchema]:
        yield from self.all_interpretations

    def interpret(self, context: ProviderContext):
        key = self.switch_on.normalize_single_value(context, self.normalization)
        interpretation_pass = self.branches.get(key, self.default)

        if interpretation_pass is not None:
            next(interpretation_pass.apply_interpretations(context))
        else:
            if self.fail_on_unhandled:
                raise UnhandledBranchError(key)



import os.path
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional, Set

from ..file_io import LoadsFromYaml, SavesToYaml
from ..pipeline import Pipeline, PipelineFile, PipelineInitializationArguments
from ..schema import ExpandsSchema, SchemaExpansionCoordinator


def get_default_name(file_path: Path) -> str:
    return file_path.stem


@dataclass
class PipelineConfiguration:
    """A `PipelineConfiguration` represents a group of configurable options for a pipeline."""

    targets: Set[str] = field(default_factory=set)
    exclude_inherited_targets: bool = False
    annotations: Dict[str, Any] = field(default_factory=dict)
    parent: Optional["PipelineConfiguration"] = None

    @classmethod
    def from_file_data(cls, data) -> "PipelineConfiguration":
        annotations = data.pop("annotations", {})
        targets = data.pop("targets", [])
        exclude_targets = data.pop("exclude_inherited_targets", False)
        return cls(targets, exclude_targets, annotations)

    @property
    def effective_targets(self) -> Set[str]:
        targets = set(self.targets)
        if not self.exclude_inherited_targets and self.parent is not None:
            targets.update(self.parent.effective_targets)
        return targets

    @property
    def effective_annotations(self) -> Dict[str, Any]:
        annotations = dict(self.parent.effective_annotations) if self.parent else {}
        annotations.update(self.annotations)
        return annotations

    def to_file_data(self, verbose: bool = False):
        annotations = self.annotations
        targets = self.targets
        result = {}
        if targets or verbose:
            result["targets"] = list(targets)
        if annotations or verbose:
            result["annotations"] = annotations
        if self.exclude_inherited_targets or verbose:
            result["exclude_inherited_targets"] = self.exclude_inherited_targets
        return result


@dataclass
class PipelineDefinition(ExpandsSchema, SavesToYaml, LoadsFromYaml):
    """A `PipelineDefinition` represents a pipeline that can be loaded from a file.

    `PipelineDefinition` objects are used to load pipelines from files. They themselves
    are not pipelines, but rather represent the metadata needed to load a pipeline from
    a file.

    `PipelineDefinition` objects are also `IntrospectiveIngestionComponent` objects,
    meaning that they can be introspected to determine their known schema definitions.

    `PipelineDefinition` objects are also `SavesToYaml` and `LoadsFromYaml` objects,
    meaning that they can be serialized to and deserialized from YAML data.
    """

    name: str
    file_path: Path
    configuration: PipelineConfiguration = field(default_factory=PipelineConfiguration)

    @classmethod
    def from_path(cls, file_path: Path):
        """Create a `PipelineDefinition` from a file path.

        The name of the pipeline will be the stem of the file path, and the annotations
        will be empty. The pipeline will be loaded from the provided file path.

        Args:
            file_path (Path): The path to the file from which to load the pipeline.

        Returns:
            PipelineDefinition: The `PipelineDefinition` object.
        """
        return cls(get_default_name(file_path), file_path)

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Or, Schema

        return Schema(
            Or(
                str,
                {
                    Optional("name"): str,
                    Optional("path"): os.path.exists,
                    Optional("annotations"): {str: Or(str, int, float, bool)},
                    Optional("targets"): [str],
                    Optional("exclude_inherited_targets"): bool,
                },
            )
        )

    @classmethod
    def from_file_data(
        cls, data, parent_configuration: PipelineConfiguration
    ) -> "PipelineDefinition":
        if isinstance(data, str):
            data = {"path": data}

        file_path = Path(data.pop("path"))
        name = data.pop("name", get_default_name(file_path))

        configuration = PipelineConfiguration.from_file_data(data)
        configuration.parent = parent_configuration
        return cls(name, file_path, configuration)

    @classmethod
    def from_plugin_data(
        cls, data, parent_configuration: PipelineConfiguration
    ) -> "PipelineDefinition":
        name = data.pop("name")
        configuration = PipelineConfiguration.from_file_data(data)
        configuration.parent = parent_configuration
        return cls(name, None, configuration)

    def to_file_data(self, verbose: bool = False):
        using_default_name = self.name == self.file_path.stem
        if using_default_name and not self.configuration.annotations and not verbose:
            return str(self.file_path)

        result = {"path": str(self.file_path)}
        if not using_default_name or verbose:
            result["name"] = self.name
        result.update(self.configuration.to_file_data(verbose))
        return result

    def to_plugin_file_data(self, verbose: bool = False):
        result = {"name": self.name}
        result.update(self.configuration.to_file_data(verbose))
        return result

    def use_configuration(self, config: PipelineConfiguration):
        self.configuration = config

    def initialize(self, init_args: PipelineInitializationArguments) -> Pipeline:
        return PipelineFile(self.file_path).load_pipeline(init_args)

    def remove_file(self, missing_ok: bool = True):
        """Remove the file associated with this `PipelineDefinition`.

        Args:
            missing_ok (bool, optional): Whether to ignore missing files. Defaults to True.

        Raises:
            FileNotFoundError: If the file does not exist and `missing_ok` is False.
        """
        self.file_path.unlink(missing_ok=missing_ok)

    def initialize_for_introspection(self) -> Pipeline:
        return self.initialize(PipelineInitializationArguments.for_introspection())

    def expand_schema(self, coordinator: SchemaExpansionCoordinator):
        self.initialize_for_introspection().expand_schema(coordinator)



from abc import ABC, abstractmethod
from typing import Dict

from ...model import PropertySet
from ...pipeline.value_providers import ProviderContext, ValueProvider


class PropertyMapping(ABC):
    @classmethod
    def from_file_data(cls, file_data):
        if isinstance(file_data, ValueProvider):
            return PropertyMappingFromValueProvider(file_data)

        providers = ValueProvider.guarantee_provider_dictionary(file_data)
        return PropertyMappingFromDict(providers)

    @abstractmethod
    def apply_to(
        self,
        context: ProviderContext,
        property_set: PropertySet,
        norm_args: Dict[str, bool],
    ):
        raise NotImplementedError


class PropertyMappingFromValueProvider(PropertyMapping):
    __slots__ = ("value_provider",)

    def __init__(self, value_provider: ValueProvider):
        self.value_provider = value_provider

    def key_value_generator(
        self, context: "ProviderContext", norm_args: Dict[str, bool]
    ):
        should_be_a_dict = self.value_provider.single_value(context)
        if not isinstance(should_be_a_dict, dict):
            raise ValueError(
                f"When using a ValueProvider as a PropertyMapping, the ValueProvider must return a dict. Instead, it returned {should_be_a_dict}"
            )

        as_providers = ValueProvider.guarantee_provider_dictionary(should_be_a_dict)

        for key, provider in as_providers.items():
            v = provider.normalize_single_value(context, norm_args)
            yield key, v

    def apply_to(
        self,
        context: ProviderContext,
        property_set: PropertySet,
        norm_args: Dict[str, bool],
    ):
        property_set.apply(self.key_value_generator(context, norm_args))

    def __iter__(self):
        # This is used when adding properties to a schema.
        # We need to know what properties are being added, so we can add them to the schema.
        # Since we cannot know them until runtime, we cannot provide them here.
        return iter([])


class PropertyMappingFromDict(PropertyMapping):
    __slots__ = ("map_of_value_providers",)

    def __init__(self, map_of_value_providers: Dict[str, ValueProvider]):
        self.map_of_value_providers = map_of_value_providers

    def key_value_generator(
        self, context: "ProviderContext", norm_args: Dict[str, bool]
    ):
        """
        Generates key-value pairs by normalizing values provided by ValueProviders.

        Args:
            context (ProviderContext): The context in which the values are being interpreted.
            norm_args (Dict[str, bool]): Arguments used for normalizing the values.

        Yields:
            Tuple[str, Any]: A tuple containing the property key and its normalized value.
        """
        for key, provider in self.map_of_value_providers.items():
            v = provider.normalize_single_value(context, norm_args)
            yield key, v

    def apply_to(
        self,
        context: ProviderContext,
        property_set: PropertySet,
        norm_args: Dict[str, bool],
    ):
        """
        Applies the key-value pairs generated by the key_value_generator to the given PropertySet.

        Args:
            context (ProviderContext): The context in which the values are being interpreted.
            property_set (PropertySet): The PropertySet to which the key-value pairs will be applied.
            norm_args (Dict[str, bool]): Arguments used for normalizing the values.
        """
        property_set.apply(self.key_value_generator(context, norm_args))

    def __iter__(self):
        return iter(self.map_of_value_providers)



from .state import (
    Adjacency,
    AdjacencyCardinality,
    Cardinality,
    ExpandsSchema,
    ExpandsSchemaFromChildren,
    GraphObjectSchema,
    GraphObjectType,
    PropertyMetadata,
    PropertyType,
    Schema,
    SchemaExpansionCoordinator,
)

__all__ = (
    "GraphObjectSchema",
    "GraphObjectType",
    "Schema",
    "ExpandsSchema",
    "ExpandsSchemaFromChildren",
    "SchemaExpansionCoordinator",
    "Cardinality",
    "AdjacencyCardinality",
    "Adjacency",
    "PropertyMetadata",
    "PropertyType",
)



from cleo.helpers import option

from ..operations import GenerateSquashedMigration
from .nodestream_command import NodestreamCommand
from .shared_options import PROJECT_FILE_OPTION


class SquashMigration(NodestreamCommand):
    name = "migrations squash"
    description = "Generate a migration for the current project."
    options = [
        PROJECT_FILE_OPTION,
        option(
            "from",
            description="The name of the migration to squash from.",
            value_required=True,
            flag=False,
        ),
        option("to", description="The name of the migration to squash to.", flag=False),
    ]

    async def handle_async(self):
        from_migration_name = self.option("from")
        to_migration_name = self.option("to")
        migrations = self.get_migrations()
        operation = GenerateSquashedMigration(
            migrations,
            from_migration_name,
            to_migration_name,
        )
        await self.run_operation(operation)



from os import environ

from nodestream.schema.state import SchemaExpansionCoordinator

from ..pipeline import Transformer
from ..pipeline.value_providers import ProviderContext
from ..schema import ExpandsSchema
from .interpretation_passes import InterpretationPass
from .interpretations import Interpretation
from .record_decomposers import RecordDecomposer

INTERPRETER_UNIQUENESS_ERROR_MESSAGE = "The interpreter must only generate source nodes in either the before_iteration phase, or the interpretations phase, not both."
VALIDATION_FLAG = "VALIDATE_PIPELINE_SCHEMA"


class InterpreterError(Exception):
    pass


class Interpreter(Transformer, ExpandsSchema):
    __slots__ = (
        "before_iteration",
        "interpretations",
        "decomposer",
    )

    @classmethod
    def from_file_data(cls, interpretations, before_iteration=None, iterate_on=None):
        # Import all interpretation plugins before we try to load any interpretations.
        Interpretation.import_all()
        interpreter_class = cls(
            before_iteration=InterpretationPass.from_file_data(before_iteration),
            interpretations=InterpretationPass.from_file_data(interpretations),
            decomposer=RecordDecomposer.from_iteration_arguments(iterate_on),
        )
        interpreter_class.verify_completeness()
        return interpreter_class

    def __init__(
        self,
        before_iteration: InterpretationPass,
        interpretations: InterpretationPass,
        decomposer: RecordDecomposer,
    ) -> None:
        self.before_iteration = before_iteration
        self.interpretations = interpretations
        self.decomposer = decomposer

    # Raise an error if both the interpretations and before_iteration phases create source nodes.
    def verify_completeness(self):
        if all(
            intepretater_phase.assigns_source_nodes
            for intepretater_phase in [self.before_iteration, self.interpretations]
        ) and environ.get(VALIDATION_FLAG, False):
            raise InterpreterError(INTERPRETER_UNIQUENESS_ERROR_MESSAGE)

    async def transform_record(self, record):
        for output_context in self.interpret_record(record):
            yield output_context.desired_ingest

    def interpret_record(self, record):
        context = ProviderContext.fresh(record)
        for base_context in self.before_iteration.apply_interpretations(context):
            for sub_context in self.decomposer.decompose_record(base_context):
                yield from self.interpretations.apply_interpretations(sub_context)

    def expand_schema(self, coordinator: SchemaExpansionCoordinator):
        # Choose between the before_iteration and interpretation passes, which one creates the source nodes.
        # Process the schema for the one that processes source nodes last so that the context of the prior is involved for each pass.
        context_pass = (
            self.before_iteration
            if self.interpretations.assigns_source_nodes
            else self.interpretations
        )
        head_pass = (
            self.interpretations
            if self.interpretations.assigns_source_nodes
            else self.before_iteration
        )

        context_pass.expand_schema(coordinator)
        head_pass.expand_schema(coordinator)



from collections import defaultdict
from typing import Dict, Iterable, List

from ...project import Project
from ...schema.migrations import ProjectMigrations
from .nodestream_command import NodestreamCommand
from .shared_options import PROJECT_FILE_OPTION, TARGETS_OPTION


class ShowMigrations(NodestreamCommand):
    name = "migrations show"
    description = (
        "List all migrations for the current project and their state on each target."
    )
    options = [PROJECT_FILE_OPTION, TARGETS_OPTION]

    def get_target_names(self, project: Project) -> List[str]:
        return self.option(TARGETS_OPTION.name) or project.targets_by_name.keys()

    async def get_migration_status_by_target(
        self,
        project: Project,
        target_names: Iterable[str],
        migrations: ProjectMigrations,
    ) -> Dict[str, Dict[str, str]]:
        migration_status_by_target = defaultdict(dict)

        for target_name in target_names:
            target = project.get_target_by_name(target_name)
            migrator = target.make_migrator()
            async for migration, pending in migrations.determine_pending(migrator):
                migration_status_by_target[migration.name][target.name] = pending

        return migration_status_by_target

    def make_migration_status_row(self, migration_name: str, statuses: Dict[str, bool]):
        def format_status(pending: bool):
            return "❌" if pending else "✅"

        return [
            migration_name,
            *[format_status(statuses[target]) for target in sorted(statuses)],
        ]

    async def generate_output_table(self):
        project = self.get_project()
        migrations = self.get_migrations()
        target_names = list(sorted(self.get_target_names(project)))
        headers = ["Migration"] + target_names
        status_by_migration = await self.get_migration_status_by_target(
            project, target_names, migrations
        )
        rows = [
            self.make_migration_status_row(
                migration_name, status_by_migration[migration_name]
            )
            for migration_name in sorted(status_by_migration)
        ]
        return headers, rows

    async def handle_async(self):
        headers, rows = await self.generate_output_table()
        table = self.table(headers, rows)
        table.render()



from typing import Any

from ...metrics import (
    AggregateHandler,
    ConsoleMetricHandler,
    JsonLogMetricHandler,
    MetricHandler,
    Metrics,
    PrometheusMetricHandler,
)
from ..commands.nodestream_command import NodestreamCommand
from ..commands.shared_options import (
    PROMETHEUS_ADDRESS_OPTION,
    PROMETHEUS_CERTFILE_OPTION,
    PROMETHEUS_KEYFILE_OPTION,
    PROMETHEUS_OPTION,
    PROMETHEUS_PORT_OPTION,
)
from .operation import Operation


class InitializeMetricsHandler(Operation):
    async def perform(self, command: NodestreamCommand) -> Any:
        Metrics.get().handler = handler = self.get_metric_handler(command)
        handler.start()

    def get_metric_handler(self, command: NodestreamCommand) -> MetricHandler:
        handlers = []
        if command.has_json_logging_set:
            handlers.append(JsonLogMetricHandler())

        prometheus_option = command.option(PROMETHEUS_OPTION.name)
        if prometheus_option:
            handlers.append(
                PrometheusMetricHandler(
                    port=int(command.option(PROMETHEUS_PORT_OPTION.name)),
                    listen_address=command.option(PROMETHEUS_ADDRESS_OPTION.name),
                    certfile=command.option(PROMETHEUS_CERTFILE_OPTION.name),
                    keyfile=command.option(PROMETHEUS_KEYFILE_OPTION.name),
                )
            )

        return AggregateHandler(handlers) if handlers else ConsoleMetricHandler(command)



from ..state import Adjacency, GraphObjectSchema, Schema
from .schema_printer import SchemaPrinter

OUTPUT_FORMAT = """Node Types:
{node_types}
Relationship Types:
{relationship_types}
Adjancies:
{adjancies}
"""

TYPE_FORMAT = "{name}: {properties}"
PROPERTY_FORMAT = "{name}: {type}"
ADJACENCY_FORMAT = "(:{from_node_type})-[:{relationship_type}]->(:{to_node_type})"


class CypherEsquePrinter(SchemaPrinter, alias="cypheresque"):
    def print_type_schema(self, node_type: GraphObjectSchema) -> str:
        formatted_properties = (
            PROPERTY_FORMAT.format(name=prop, type=meta.type.value)
            for prop, meta in node_type.properties.items()
        )

        return TYPE_FORMAT.format(
            name=node_type.name, properties=", ".join(formatted_properties)
        )

    def join_line_seperated(self, lines: list) -> str:
        return "\n".join(lines)

    def print_adjanceny(self, adjancency: Adjacency) -> str:
        return ADJACENCY_FORMAT.format(
            from_node_type=adjancency.from_node_type,
            relationship_type=adjancency.relationship_type,
            to_node_type=adjancency.to_node_type,
        )

    def print_node_types(self, schema: Schema) -> str:
        return self.join_line_seperated(
            self.print_type_schema(node_type) for node_type in schema.nodes
        )

    def print_relationship_types(self, schema: Schema) -> str:
        return self.join_line_seperated(
            self.print_type_schema(relationship_type)
            for relationship_type in schema.relationships
        )

    def print_adjancies(self, schema: Schema) -> str:
        return self.join_line_seperated(
            self.print_adjanceny(adjancency) for adjancency in schema.adjacencies
        )

    def print_schema_to_string(self, schema: Schema) -> str:
        return OUTPUT_FORMAT.format(
            node_types=self.print_node_types(schema),
            relationship_types=self.print_relationship_types(schema),
            adjancies=self.print_adjancies(schema),
        )



from abc import ABC, abstractmethod
from contextlib import contextmanager
from contextvars import ContextVar
from enum import Enum, auto
from logging import getLogger
from typing import Optional, Union

from cleo.commands.command import Command

Number = Union[int, float]


class Metric(Enum):
    RECORDS = auto()
    NODES_UPSERTED = auto()
    RELATIONSHIPS_UPSERTED = auto()
    TIME_TO_LIVE_OPERATIONS = auto()
    INGEST_HOOKS_EXECUTED = auto()
    NON_FATAL_ERRORS = auto()
    STEPS_RUNNING = auto()
    BUFFERED_RECORDS = auto()


class MetricHandler(ABC):
    """Recieves metrics and handles them in some way.

    A metric handler is a class that recieves metrics and handles them in some way.
    """

    def start(self):
        """Start the handler.

        This method is called before any metrics are sent to the handler.
        It can be used to initialize resources or start background threads.
        """
        pass

    def stop(self):
        """Stop the handler.

        This method is called after all metrics have been sent to the handler.
        It can be used to release resources or stop background threads.
        """
        pass

    @abstractmethod
    def increment(self, metric: Metric, value: Number): ...

    @abstractmethod
    def decrement(self, metric: Metric, value: Number): ...


class NullMetricHandler(MetricHandler):
    """A metric handler that does nothing."""

    def increment(self, _: Metric, __: Number):
        pass

    def decrement(self, _: Metric, __: Number):
        pass


RECORDS_COUNTER_NAME = "records"
NON_FATAL_ERRORS_COUNTER_NAME = "non_fatal_errors"
NODES_UPSERTED_COUNTER_NAME = "nodes_upserted"
RELS_UPSERTED_COUNTER_NAME = "relationships_upserted"
TTL_OPERATIONS_COUNTER_NAME = "time_to_live_operations"
INGEST_HOOKS_EXECUTED_COUNTER_NAME = "ingest_hooks_executed"
BUFFERED_RECORDS_COUNTER_NAME = "buffered_records"

RECORDS_COUNTER_DESCRIPTION = "Number of records processed"
NON_FATAL_ERRORS_COUNTER_DESCRIPTION = "Number of non-fatal errors"
NODES_UPSERTED_COUNTER_DESCRIPTION = "Number of nodes upserted to the graph"
RELS_UPSERTED_COUNTER_DESCRIPTION = "Number of relationships upserted to the graph"
TTL_OPERATIONS_COUNTER_DESCRIPTION = "Number of time-to-live operations executed"
INGEST_HOOKS_EXECUTED_COUNTER_DESCRIPTION = (
    "Number of ingest hooks executed to the graph"
)
STEPS_RUNNING_COUNTER_DESCRIPTION = "Number of steps currently running in the pipeline"
BUFFERED_RECORDS_DESCRIPTION = "Number of records buffered between steps"


try:

    from prometheus_client import REGISTRY, Gauge, start_http_server

    class PrometheusMetricHandler(MetricHandler):
        """A metric handler that sends metrics to Prometheus."""

        def __init__(
            self,
            port: int = 9090,
            listen_address: str = "0.0.0.0",
            certfile: Optional[str] = None,
            keyfile: Optional[str] = None,
        ):
            self.logger = getLogger(__name__)
            self.instruments_by_metric = {
                Metric.RECORDS: Gauge(
                    name=RECORDS_COUNTER_NAME,
                    documentation=RECORDS_COUNTER_DESCRIPTION,
                ),
                Metric.NON_FATAL_ERRORS: Gauge(
                    name=NON_FATAL_ERRORS_COUNTER_NAME,
                    documentation=NON_FATAL_ERRORS_COUNTER_DESCRIPTION,
                ),
                Metric.NODES_UPSERTED: Gauge(
                    name=NODES_UPSERTED_COUNTER_NAME,
                    documentation=NODES_UPSERTED_COUNTER_DESCRIPTION,
                ),
                Metric.RELATIONSHIPS_UPSERTED: Gauge(
                    name=RELS_UPSERTED_COUNTER_NAME,
                    documentation=RELS_UPSERTED_COUNTER_DESCRIPTION,
                ),
                Metric.TIME_TO_LIVE_OPERATIONS: Gauge(
                    name=TTL_OPERATIONS_COUNTER_NAME,
                    documentation=TTL_OPERATIONS_COUNTER_DESCRIPTION,
                ),
                Metric.INGEST_HOOKS_EXECUTED: Gauge(
                    name=INGEST_HOOKS_EXECUTED_COUNTER_NAME,
                    documentation=INGEST_HOOKS_EXECUTED_COUNTER_DESCRIPTION,
                ),
                Metric.STEPS_RUNNING: Gauge(
                    name="steps_running",
                    documentation=STEPS_RUNNING_COUNTER_DESCRIPTION,
                ),
                Metric.BUFFERED_RECORDS: Gauge(
                    name=BUFFERED_RECORDS_COUNTER_NAME,
                    documentation=BUFFERED_RECORDS_DESCRIPTION,
                ),
            }

            self.port = port
            self.listen_address = listen_address
            self.certfile = certfile
            self.keyfile = keyfile

        def start(self):
            server, thread = start_http_server(
                port=self.port,
                addr=self.listen_address,
                certfile=self.certfile,
                keyfile=self.keyfile,
            )
            self.server = server
            self.thread = thread
            self.logger.info(
                f"Prometheus metrics server started on {self.listen_address}:{self.port}"
            )

        def stop(self):
            for instrument in self.instruments_by_metric.values():
                REGISTRY.unregister(instrument)
            self.logger.info("Shutting down prometheus metrics server")
            self.server.shutdown()
            self.logger.debug("Waiting for prometheus metrics server to shut down")
            self.thread.join()
            self.logger.info("Prometheus metrics server shut down successfully")

        def increment(self, metric: Metric, value: Number):
            self.instruments_by_metric[metric].inc(value)

        def decrement(self, metric: Metric, value: Number):
            self.instruments_by_metric[metric].dec(value)

except ImportError:

    class PrometheusMetricHandler(MetricHandler):
        """A metric handler that sends metrics to Prometheus."""

        def __init__(self, *args, **kwargs):
            raise ImportError(
                "The prometheus_client library is required to use the PrometheusMetricHandler. "
                "Please install the `prometheus` extra."
            )

        def increment(self, metric: Metric, value: Number):
            pass

        def decrement(self, metric: Metric, value: Number):
            pass


STATS_TABLE_COLS = ["Statistic", "Value"]


class ConsoleMetricHandler(MetricHandler):
    """A metric handler that prints metrics to the console."""

    def __init__(self, command: Command):
        self.metrics = {x: 0 for x in Metric}
        self.command = command

    def increment(self, metric: Metric, value: Number):
        self.metrics[metric] += value

    def decrement(self, metric: Metric, value: Number):
        self.metrics[metric] -= value

    def stop(self):
        stats = ((k.name, str(v)) for k, v in self.metrics.items() if v > 0)
        table = self.command.table(STATS_TABLE_COLS, stats)
        table.render()


class JsonLogMetricHandler(MetricHandler):
    """A metric handler that logs metrics in JSON format."""

    def __init__(self):
        self.metrics = {x: 0 for x in Metric}
        self.logger = getLogger(__name__)

    def increment(self, metric: Metric, value: Number):
        self.metrics[metric] += value

    def decrement(self, metric: Metric, value: Number):
        self.metrics[metric] -= value

    def stop(self):
        self.logger.info(
            "Metrics Report", extra={k.name: v for k, v in self.metrics.items()}
        )


class AggregateHandler(MetricHandler):
    """A metric handler that aggregates metrics from multiple handlers."""

    def __init__(self, handlers):
        self.handlers = handlers

    def start(self):
        for handler in self.handlers:
            handler.start()

    def stop(self):
        for handler in self.handlers:
            handler.stop()

    def increment(self, metric: Metric, value: Number):
        for handler in self.handlers:
            handler.increment(metric, value)

    def decrement(self, metric: Metric, value: Number):
        for handler in self.handlers:
            handler.decrement(metric, value)


UNKNOWN_PIPELINE_NAME = "unknown"
UNKNOWN_PIPELINE_SCOPE = "unknown"
context: ContextVar["Metrics"] = ContextVar("context")


class Metrics:
    def __init__(
        self,
        handler: Optional[MetricHandler] = None,
        scope_name: Optional[str] = None,
        pipeline_name: Optional[str] = None,
    ):
        self.handler = handler or NullMetricHandler()
        self.scope_name = scope_name or UNKNOWN_PIPELINE_SCOPE
        self.pipeline_name = pipeline_name or UNKNOWN_PIPELINE_NAME

    def increment(self, metric: Metric, value: Number = 1):
        self.handler.increment(metric, value)

    def decrement(self, metric: Metric, value: Number = 1):
        self.handler.decrement(metric, value)

    def start(self):
        self.handler.start()

    def stop(self):
        self.handler.stop()

    @staticmethod
    @contextmanager
    def capture(handler: Optional[MetricHandler] = None):
        metrics = Metrics(handler)
        metrics.start()
        token = context.set(metrics)
        try:
            yield metrics
        finally:
            context.reset(token)
            metrics.stop()

    @staticmethod
    def get() -> "Metrics":
        try:
            return context.get()
        except LookupError:
            return Metrics()

    def contextualize(self, scope_name: str, pipeline_name: str):
        self.scope_name = scope_name
        self.pipeline_name = pipeline_name



from typing import Dict, Iterable, List, Optional, Set, Tuple

from ..state import GraphObjectSchema
from .operations import (
    AddAdditionalNodePropertyIndex,
    AddAdditionalRelationshipPropertyIndex,
    AddNodeProperty,
    AddRelationshipProperty,
    CreateNodeType,
    CreateRelationshipType,
    DropAdditionalNodePropertyIndex,
    DropAdditionalRelationshipPropertyIndex,
    DropNodeProperty,
    DropNodeType,
    DropRelationshipProperty,
    DropRelationshipType,
    NodeKeyExtended,
    NodeKeyPartRenamed,
    Operation,
    RelationshipKeyExtended,
    RelationshipKeyPartRenamed,
    RenameNodeProperty,
    RenameNodeType,
    RenameRelationshipProperty,
    RenameRelationshipType,
)
from .state_providers import StateProvider


class MigratorInput:
    """Asks questions as the schema change detector works.

    This class is used by the schema change detector to ask questions to the
    user as it works. The schema change detector will ask questions to the
    user when it is not sure about the answer.

    The default implementation of this class will always answer no to all
    questions.
    """

    def ask_yes_no(self, question: str) -> bool:
        """Ask a yes/no question.

        Args:
            question: The question to ask.

        Returns:
            True if the user answered yes, False if the user answered no.
        """
        return False

    def format_ask_type_renamed(self, old_type: str, new_type: str) -> str:
        """Format a question about a type being renamed.

        Args:
            old_type: The old type name.
            new_type: The new type name.

        Returns:
            The formatted question.
        """
        return f"Did you rename {old_type} to {new_type}?"

    def ask_type_renamed(self, old_type: str, new_type: str) -> bool:
        """Ask if a type was renamed.

        Args:
            old_type: The old type name.
            new_type: The new type name.

        Returns:
            True if the user answered yes, False if the user answered no.
        """
        question = self.format_ask_type_renamed(old_type, new_type)
        return self.ask_yes_no(question)

    def format_ask_property_renamed(
        self, object_type: str, old_property_name: str, new_property_name: str
    ) -> str:
        """Format a question about a property being renamed.

        Args:
            object_type: The type of the object that the property belongs to.
            old_property_name: The old property name.
            new_property_name: The new property name.

        Returns:
            The formatted question.
        """
        return f"Did you rename {old_property_name} to {new_property_name} on {object_type}?"

    def ask_property_renamed(
        self, object_type: str, old_property_name: str, new_property_name: str
    ) -> bool:
        """Ask if a property was renamed.

        Args:
            object_type: The type of the object that the property belongs to.
            old_property_name: The old property name.
            new_property_name: The new property name.

        Returns:
            True if the user answered yes, False if the user answered no.
        """
        question = self.format_ask_property_renamed(
            object_type, old_property_name, new_property_name
        )
        return self.ask_yes_no(question)

    def determine_renamed_types(
        self, presumed_added_types: Set[str], presumed_deleted_types: Set[str]
    ) -> Set[Tuple[str, str]]:
        """Determine which types have been renamed.

        Args:
            presumed_added_types: The types that are presumed to have been
                added which may have been renamed.
            presumed_deleted_types: The types that are presumed to have been
                deleted which may have been renamed.

        Returns:
            A set of tuples containing the old and new type names.
        """
        renamed = set()

        for new_type_name in presumed_added_types:
            for prev_type_name in presumed_deleted_types:
                if self.ask_type_renamed(prev_type_name, new_type_name):
                    renamed.add((prev_type_name, new_type_name))
                    break

        return renamed

    def determine_renamed_properties(
        self,
        object_type: str,
        presumed_added_properties: Set[str],
        presumed_deleted_properties: Set[str],
    ) -> Set[Tuple[str, str]]:
        """Determine which properties have been renamed.

        Args:
            presumed_added_properties: The properties that are presumed to have been
                added which may have been renamed.
            presumed_deleted_properties: The properties that are presumed to have been
                deleted which may have been renamed.

        Returns:
            A set of tuples containing the old and new property names.
        """
        renamed = set()

        for new_property_name in presumed_added_properties:
            for prev_property_name in presumed_deleted_properties:
                if self.ask_property_renamed(
                    object_type, prev_property_name, new_property_name
                ):
                    renamed.add((prev_property_name, new_property_name))
                    break

        return renamed


class TypePairing:
    """A pair of types from two different states.

    This class represents a pair of types from two different states. It
    is used to compare the two types and determine what has changed.
    """

    __slots__ = ("from_type", "to_type")

    def __init__(
        self,
        from_type: Optional[GraphObjectSchema],
        to_type: Optional[GraphObjectSchema],
    ) -> None:
        """Initialize the type bridge.

        Args:
            from_type: The name of the type before the migration.
            to_type: The name of the type after the migration.
        """
        self.from_type = from_type or GraphObjectSchema(name="")
        self.to_type = to_type or GraphObjectSchema(name="")

    def get_property_drift(
        self, input: MigratorInput
    ) -> Tuple[Set[str], Set[str], Set[Tuple[str, str]]]:
        """Get the property drift between the two types.

        Returns:
            A tuple containing the set of properties that were deleted, the
            set of properties that were added, and the set of properties that
            were renamed.
        """
        deleted = self.from_type.non_key_properties - self.to_type.non_key_properties
        added = self.to_type.non_key_properties - self.from_type.non_key_properties
        renamed = input.determine_renamed_properties(self.to_type.name, added, deleted)
        deleted = deleted - {item[0] for item in renamed}
        added = added - {item[1] for item in renamed}
        return deleted, added, renamed

    def get_key_drift(
        self, input: MigratorInput
    ) -> Tuple[Set[str], Set[str], Set[Tuple[str, str]]]:
        """Get the key drift between the two types.

        Returns:
            A tuple containing the set of keys that were deleted, the
            set of keys that were added, and the set of keys that
            were renamed.
        """
        deleted = self.from_type.keys - self.to_type.keys
        added = self.to_type.keys - self.from_type.keys
        renamed = input.determine_renamed_properties(self.to_type.name, added, deleted)
        deleted = deleted - {item[0] for item in renamed}
        added = added - {item[1] for item in renamed}
        return deleted, added, renamed

    def get_index_drift(self) -> Tuple[Set[str], Set[str]]:
        """Get the index drift between the two types.

        Returns:
            A tuple containing the set of indexed properties that were deleted,
            the set of indexed properties that were added, and the set of indexed
            properties that were renamed.
        """
        deleted = self.from_type.indexed_properties - self.to_type.indexed_properties
        added = self.to_type.indexed_properties - self.from_type.indexed_properties
        return deleted, added


class AutoChangeDetector:
    """Detects changes between two schemas.

    Detects changes between two schemas and produces a migration that
    can be used to migrate the database from the first schema to the
    second schema.
    """

    __slots__ = (
        "input",
        "from_state_provider",
        "to_state_provider",
        "from_state",
        "to_state",
        "new_node_types",
        "renamed_node_types",
        "deleted_node_types",
        "new_relationship_types",
        "renamed_relationship_types",
        "deleted_relationship_types",
        "deleted_relationship_properties",
        "renamed_relationship_properties",
        "added_relationship_properties",
        "deleted_node_properties",
        "renamed_node_properties",
        "added_node_properties",
        "deleted_relationship_keys",
        "renamed_relationship_keys",
        "added_relationship_keys",
        "deleted_node_keys",
        "renamed_node_keys",
        "added_node_keys",
        "added_node_property_indexes",
        "deleted_node_property_indexes",
        "added_relationship_property_indexes",
        "deleted_relationship_property_indexes",
    )

    def __init__(
        self,
        input: MigratorInput,
        from_state: StateProvider,
        to_state: StateProvider,
    ) -> None:
        """Initialize the autodetector.

        Args:
            input: Input for asking questions when detecting ambiguous changes.
            from_state: The state provider for the state before the migration.
            to_state: The state provider for the state after the migration.
        """
        self.input = input
        self.from_state_provider = from_state
        self.to_state_provider = to_state
        # Node Types
        self.new_node_types = set()
        self.renamed_node_types = set()
        self.deleted_node_types = set()
        # Relationship Types
        self.new_relationship_types = set()
        self.renamed_relationship_types = set()
        self.deleted_relationship_types = set()
        # Relationship Properties
        self.deleted_relationship_properties = set()
        self.renamed_relationship_properties = set()
        self.added_relationship_properties = set()
        # Node Properties
        self.deleted_node_properties = set()
        self.renamed_node_properties = set()
        self.added_node_properties = set()
        # Relationship Keys
        self.deleted_relationship_keys = set()
        self.renamed_relationship_keys = set()
        self.added_relationship_keys = set()
        # Node Keys
        self.deleted_node_keys = set()
        self.renamed_node_keys = set()
        self.added_node_keys = set()
        # Node Indexes
        self.added_node_property_indexes = set()
        self.deleted_node_property_indexes = set()
        # Relationship Indexes
        self.added_relationship_property_indexes = set()
        self.deleted_relationship_property_indexes = set()

    async def get_to_and_from_state(self):
        """Get the to and from state from the state providers."""
        self.from_state = await self.from_state_provider.get_schema()
        self.to_state = await self.to_state_provider.get_schema()

    async def detect_changes(self) -> List[Operation]:
        """Detect changes between the two states and return operations.

        Returns:
            A list of operations that can be used to migrate the database
            from the from_state to the to_state.
        """
        await self.get_to_and_from_state()
        self.detect_node_type_changes()
        self.detect_relationship_type_changes()
        self.detect_node_property_changes()
        self.detect_node_key_changes()
        self.detect_relationship_property_changes()
        self.detect_node_index_changes()
        self.detect_relationship_index_changes()
        self.detect_relationship_key_changes()

        operations = []
        # Update Nodes
        operations.extend(self.make_node_type_change_operations())
        operations.extend(self.make_node_property_change_operations())
        operations.extend(self.make_node_index_change_operations())
        operations.extend(self.make_node_key_change_operations())

        # Update Relationships
        operations.extend(self.make_relationship_type_change_operations())
        operations.extend(self.make_relationship_property_change_operations())
        operations.extend(self.make_relationship_index_change_operations())
        operations.extend(self.make_relationship_key_change_operations())

        return operations

    def make_node_key_change_operations(self) -> Iterable[Operation]:
        if self.deleted_node_keys:
            raise NotImplementedError("Deleting node keys is not supported yet.")

        for type, old_field, new_field in self.renamed_node_keys:
            yield NodeKeyPartRenamed(type, old_field, new_field)

        for type, new_field in self.added_node_keys:
            yield NodeKeyExtended(type, new_field, None)

    def make_relationship_key_change_operations(self) -> Iterable[Operation]:
        if self.deleted_relationship_keys:
            raise NotImplementedError(
                "Deleting relationship keys is not supported yet."
            )

        for type, old_field, new_field in self.renamed_relationship_keys:
            yield RelationshipKeyPartRenamed(type, old_field, new_field)

        for type, new_field in self.added_relationship_keys:
            yield RelationshipKeyExtended(type, new_field, None)

    def make_node_property_change_operations(self) -> Iterable[Operation]:
        for type, deleted_field in self.deleted_node_properties:
            if (type, deleted_field) in self.added_node_keys:
                continue
            yield DropNodeProperty(type, deleted_field)

        for type, old_field, new_field in self.renamed_node_properties:
            yield RenameNodeProperty(type, old_field, new_field)

        for type, new_field in self.added_node_properties:
            yield AddNodeProperty(type, new_field)

    def make_relationship_property_change_operations(self) -> Iterable[Operation]:
        for type, deleted_field in self.deleted_relationship_properties:
            if (type, deleted_field) in self.added_relationship_keys:
                continue
            yield DropRelationshipProperty(type, deleted_field)

        for type, old_field, new_field in self.renamed_relationship_properties:
            yield RenameRelationshipProperty(type, old_field, new_field)

        for type, new_field in self.added_relationship_properties:
            yield AddRelationshipProperty(type, new_field)

    def make_node_type_change_operations(self) -> Iterable[Operation]:
        for deleted_type in self.deleted_node_types:
            yield DropNodeType(deleted_type)

        for old_type, new_type in self.renamed_node_types:
            yield RenameNodeType(old_type, new_type)

        for new_type in self.new_node_types:
            type_def = self.to_state.get_node_type_by_name(new_type)
            yield CreateNodeType(
                name=new_type,
                keys=type_def.keys,
                properties=type_def.non_key_properties,
            )

    def make_relationship_type_change_operations(self) -> Iterable[Operation]:
        for deleted_type in self.deleted_relationship_types:
            yield DropRelationshipType(deleted_type)

        for old_type, new_type in self.renamed_relationship_types:
            yield RenameRelationshipType(old_type, new_type)

        for new_type in self.new_relationship_types:
            type_def = self.to_state.get_relationship_type_by_name(new_type)
            yield CreateRelationshipType(
                name=new_type,
                keys=type_def.keys,
                properties=type_def.non_key_properties,
            )

    def make_relationship_index_change_operations(self) -> Iterable[Operation]:
        for deleted_type, deleted_field in self.deleted_relationship_property_indexes:
            yield DropAdditionalRelationshipPropertyIndex(deleted_type, deleted_field)

        for new_type, new_field in self.added_relationship_property_indexes:
            yield AddAdditionalRelationshipPropertyIndex(new_type, new_field)

    def make_node_index_change_operations(self) -> Iterable[Operation]:
        for deleted_type, deleted_field in self.deleted_node_property_indexes:
            yield DropAdditionalNodePropertyIndex(deleted_type, deleted_field)

        for new_type, new_field in self.added_node_property_indexes:
            yield AddAdditionalNodePropertyIndex(new_type, new_field)

    def detect_node_key_changes(self):
        for pair in self.get_node_pairs():
            deleted_keys, added_keys, renamed_keys = pair.get_key_drift(self.input)
            for key in deleted_keys:
                self.deleted_node_keys.add((pair.to_type.name, key))
            for old, new in renamed_keys:
                self.renamed_node_keys.add((pair.to_type.name, old, new))
            for key in added_keys:
                self.added_node_keys.add((pair.to_type.name, key))

    def detect_relationship_key_changes(self):
        for pair in self.get_relationship_pairs():
            deleted_keys, added_keys, renamed_keys = pair.get_key_drift(self.input)
            for key in deleted_keys:
                self.deleted_relationship_keys.add((pair.to_type.name, key))
            for old, new in renamed_keys:
                self.renamed_relationship_keys.add((pair.to_type.name, old, new))
            for key in added_keys:
                self.added_relationship_keys.add((pair.to_type.name, key))

    def detect_node_type_changes(self):
        self.deleted_node_types, self.new_node_types = self.from_state.diff_node_types(
            self.to_state
        )

        self.renamed_node_types = self.input.determine_renamed_types(
            self.new_node_types, self.deleted_node_types
        )

        for previous, new in self.renamed_node_types:
            self.deleted_node_types.remove(previous)
            self.new_node_types.remove(new)

    def detect_relationship_type_changes(self):
        (
            self.deleted_relationship_types,
            self.new_relationship_types,
        ) = self.from_state.diff_relationship_types(self.to_state)

        self.renamed_relationship_types = self.input.determine_renamed_types(
            self.new_relationship_types, self.deleted_relationship_types
        )

        for previous, new in self.renamed_relationship_types:
            self.deleted_relationship_types.remove(previous)
            self.new_relationship_types.remove(new)

    def detect_node_property_changes(self):
        for pair in self.get_node_pairs():
            (
                deleted_properties,
                added_properties,
                renamed_properties,
            ) = pair.get_property_drift(self.input)
            for prop in deleted_properties:
                self.deleted_node_properties.add((pair.to_type.name, prop))
            for prop in added_properties:
                self.added_node_properties.add((pair.to_type.name, prop))
            for old, new in renamed_properties:
                self.renamed_node_properties.add((pair.to_type.name, old, new))

    def detect_relationship_property_changes(self):
        for pair in self.get_relationship_pairs():
            (
                deleted_properties,
                added_properties,
                renamed_properties,
            ) = pair.get_property_drift(self.input)
            for prop in deleted_properties:
                self.deleted_relationship_properties.add((pair.to_type.name, prop))
            for prop in added_properties:
                self.added_relationship_properties.add((pair.to_type.name, prop))
            for old, new in renamed_properties:
                self.renamed_relationship_properties.add((pair.to_type.name, old, new))

    def detect_node_index_changes(self):
        for pair in self.get_node_pairs(ignore_created_and_deleted=False):
            deleted_indexes, added_indexes = pair.get_index_drift()
            self.deleted_node_property_indexes.update(
                (pair.to_type.name, idx) for idx in deleted_indexes
            )
            self.added_node_property_indexes.update(
                (pair.to_type.name, idx) for idx in added_indexes
            )

    def detect_relationship_index_changes(self):
        for pair in self.get_relationship_pairs(ignore_created_and_deleted=False):
            deleted_indexes, added_indexes = pair.get_index_drift()
            self.deleted_relationship_property_indexes.update(
                (pair.to_type.name, idx) for idx in deleted_indexes
            )
            self.added_relationship_property_indexes.update(
                (pair.to_type.name, idx) for idx in added_indexes
            )

    def get_node_pairs(
        self, ignore_created_and_deleted: bool = True
    ) -> Iterable[TypePairing]:
        return self.compare_types_pairwise(
            self.from_state.nodes_by_name,
            self.to_state.nodes_by_name,
            self.deleted_node_types,
            self.new_node_types,
            self.renamed_node_types,
            ignore_created_and_deleted,
        )

    def get_relationship_pairs(
        self, ignore_created_and_deleted: bool = True
    ) -> Iterable[TypePairing]:
        return self.compare_types_pairwise(
            self.from_state.relationships_by_name,
            self.to_state.relationships_by_name,
            self.deleted_relationship_types,
            self.new_relationship_types,
            self.renamed_relationship_types,
            ignore_created_and_deleted,
        )

    def compare_types_pairwise(
        self,
        from_population: Dict[str, GraphObjectSchema],
        to_population: Dict[str, GraphObjectSchema],
        deleted_types: Set[str],
        new_types: Set[str],
        renamed_types: Set[Tuple[str, str]],
        ignore_created_and_deleted: bool,
    ) -> Iterable[TypePairing]:
        all_types = set(from_population).union(to_population)
        for type in all_types:
            # If the type is new or deleted, we can ignore it.
            new_or_removed = type in new_types or type in deleted_types
            if new_or_removed and ignore_created_and_deleted:
                continue

            # If the type has been renamed, we can compare the old an
            # new type definitions. We only want to consider the type once,
            # so we'll ignore it if type is the old value.
            renamed_from = next(
                (old for old, new in renamed_types if new == type),
                None,
            )
            renamed_to = next(
                (new for old, new in renamed_types if old == type),
                None,
            )
            if renamed_to:
                continue

            # If the type was renamed, we can inform the property
            # diffing logic the two types. If it wasn't, it can be the
            # same name.
            if renamed_from:
                yield TypePairing(
                    from_type=from_population.get(renamed_from),
                    to_type=to_population.get(type),
                )
            else:
                yield TypePairing(
                    from_type=from_population.get(type),
                    to_type=to_population.get(type),
                )



from ..state import Schema
from .schema_printer import SchemaPrinter


class PlainTestSchemaPrinter(SchemaPrinter, alias="plain"):
    def print_schema_to_string(self, schema: Schema):
        return f"{schema}\n"



from cleo.helpers import argument, option

PROJECT_FILE_OPTION = option(
    "project", "p", "The project file (nodestream.yaml) to load.", flag=False
)

DATABASE_NAME_OPTION = option(
    "database", "d", "The Database to Configure", flag=False, default="neo4j"
)

SCOPE_NAME_OPTION = option(
    "scope", "s", "Use the provided scope for the operation", flag=False
)

JSON_OPTION = option("json", "j", "Log output in JSON", flag=True)


PIPELINE_ARGUMENT = argument("pipeline", "the name of the pipeline")
MANY_PIPELINES_ARGUMENT = argument(
    "pipelines", "the names of the pipelines", multiple=True, optional=True
)

TARGETS_OPTION = option(
    "target",
    "t",
    "Specify a database to target at run time.",
    multiple=True,
    flag=False,
)

PROMETHEUS_OPTION = option(
    "prometheus",
    description="Enable the Prometheus metrics server",
    flag=True,
)

PROMETHEUS_PORT_OPTION = option(
    "prometheus-listen-port",
    description="The port to run the Prometheus metrics server on",
    flag=False,
    default="9090",
)

PROMETHEUS_ADDRESS_OPTION = option(
    "prometheus-listen-address",
    description="The address to run the Prometheus metrics server on",
    flag=False,
    default="0.0.0.0",
)

PROMETHEUS_CERTFILE_OPTION = option(
    "prometheus-certfile",
    description="The path to the certificate file for the Prometheus metrics server",
    flag=False,
)

PROMETHEUS_KEYFILE_OPTION = option(
    "prometheus-keyfile",
    description="The path to the key file for the Prometheus metrics server",
    flag=False,
)


PROMETHEUS_OPTIONS = [
    PROMETHEUS_OPTION,
    PROMETHEUS_PORT_OPTION,
    PROMETHEUS_ADDRESS_OPTION,
    PROMETHEUS_CERTFILE_OPTION,
    PROMETHEUS_KEYFILE_OPTION,
]



from abc import ABC, abstractmethod

from ...pipeline.value_providers import ProviderContext
from ...pluggable import Pluggable
from ...schema import ExpandsSchema
from ...subclass_registry import SubclassRegistry
from .conditions import Condition

INTERPRETATION_REGISTRY = SubclassRegistry()


@INTERPRETATION_REGISTRY.connect_baseclass
class Interpretation(ExpandsSchema, Pluggable, ABC):
    entrypoint_name = "interpretations"
    assigns_source_nodes = False

    @abstractmethod
    def interpret(self, context: ProviderContext):
        raise NotImplementedError

    @classmethod
    def from_file_data(cls, **arguments) -> "Interpretation":
        name = arguments.pop("type")
        condition = arguments.pop("condition", None)
        class_to_load = INTERPRETATION_REGISTRY.get(name)
        interpretation = class_to_load(**arguments)
        if condition:
            condition = Condition.from_tagged_file_data(**condition)
            return ConditionedInterpretation(condition, interpretation)
        return interpretation


class ConditionedInterpretation(Interpretation):
    def __init__(self, condition: Condition, interpretation: Interpretation):
        self.condition = condition
        self.interpretation = interpretation

    def interpret(self, context: ProviderContext):
        if self.condition.evaluate(context):
            self.interpretation.interpret(context)



from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from ..pipeline.object_storage import ObjectStore, Signer


@dataclass(frozen=True, slots=True)
class StoreConfiguration:
    name: str
    storage_type: str
    arguments: Dict[str, Any]
    hmac_key: Optional[any] = None

    def initialize(self) -> ObjectStore:
        store = ObjectStore.from_file_arguments(self.storage_type, **self.arguments)
        if self.hmac_key:
            return store.signed(Signer.hmac(self.hmac_key))
        else:
            return store

    def to_file_data(self):
        return dict(
            name=self.name,
            type=self.storage_type,
            hmac_key=self.hmac_key,
            **self.arguments
        )

    @staticmethod
    def from_file_data(data):
        storage_type = data.pop("type")
        name = data.pop("name")
        hmac_key = data.pop("hmac_key", None)
        return StoreConfiguration(
            name=name, storage_type=storage_type, hmac_key=hmac_key, arguments=data
        )

    @staticmethod
    def describe_yaml_schema():
        from schema import Optional, Schema

        return Schema(
            {"name": str, "type": str, Optional("hmac_key"): str, Optional(str): object}
        )


@dataclass(frozen=True, slots=True)
class StorageConfiguration:
    storage_configuration_by_name: Dict[str, StoreConfiguration] = field(
        default_factory=dict
    )

    def initialize_by_name(self, name: str) -> ObjectStore:
        if name not in self.storage_configuration_by_name:
            return ObjectStore.null()

        return self.storage_configuration_by_name[name].initialize()

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            {
                Optional("stores"): [
                    StoreConfiguration.describe_yaml_schema(),
                ]
            }
        )

    @classmethod
    def from_file_data(cls, data):
        storage_configuration_by_name = {
            store.name: store
            for store in (
                StoreConfiguration.from_file_data(store_data)
                for store_data in data.get("stores", [])
            )
        }

        return StorageConfiguration(storage_configuration_by_name)

    def to_file_data(self) -> dict:
        return {"stores": [v for v in self.storage_configuration_by_name.items()]}
